SystemOrganization addCategory: #'System-Text'!

Editor subclass: #TextEditor
	instanceVariableNames: 'model editingState'
	classVariableNames: 'CaseSensitiveFinds UseFindReplaceSelection UseSecondarySelection UseSelectionBar WalkAlongDisplayedLine'
	poolDictionaries: ''
	category: 'System-Text'!

TextEditor class
	instanceVariableNames: 'cmdActions shiftCmdActions'!

!TextEditor commentStamp: '<historical>' prior: 0!
See comment in Editor.

My instances edit Text, this is, they support multiple lines and TextAttributes.
They have no specific facilities for editing Smalltalk code. Those are found in SmalltalkEditor.!

TextEditor class
	instanceVariableNames: 'cmdActions shiftCmdActions'!

TextEditor subclass: #SmalltalkEditor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Text'!

!SmalltalkEditor commentStamp: '<historical>' prior: 0!
The editor built specifically for Smalltalk code!

!SmalltalkEditor class methodsFor: 'shortcuts' stamp: 'GuillermoPolito 5/26/2012 21:14'!
buildSmalltalkEditorCtrlBasedKeymappingsOn: aBuilder
	(aBuilder shortcut: #browseIt1) 
		category: #SmalltalkEditor
		default: $b command
		do: [ :morph | morph browseIt ].
	
	(aBuilder shortcut: #browseIt2) 
		category: #SmalltalkEditor
		default: $b ctrl
		do: [ :morph | morph browseIt ].
				
	(aBuilder shortcut: #doIt1) 
		category: #SmalltalkEditor
		default: $d command
		do: [ :morph | morph doIt ].
	
	(aBuilder shortcut: #doIt2) 
		category: #SmalltalkEditor
		default: $d ctrl
		do: [ :morph | morph doIt ].
		
	(aBuilder shortcut: #inspectIt1) 
		category: #SmalltalkEditor
		default: $i command
		do: [ :morph | morph inspectIt ].
	
	(aBuilder shortcut: #inspectIt2) 
		category: #SmalltalkEditor
		default: $i ctrl
		do: [ :morph | morph inspectIt ].
		
	(aBuilder shortcut: #implementorsOfIt1) 
		category: #SmalltalkEditor
		default: $m command
		do: [ :morph | morph implementorsOfIt ].

	(aBuilder shortcut: #implementorsOfIt2) 
		category: #SmalltalkEditor
		default: $m ctrl
		do: [ :morph | morph implementorsOfIt ].
		
	(aBuilder shortcut: #sendersOfIt1)
		category: #SmalltalkEditor
		default: $n command
		do: [ :morph | morph sendersOfIt ].

	(aBuilder shortcut: #sendersOfIt2) 
		category: #SmalltalkEditor
		default: $n ctrl
		do: [ :morph | morph sendersOfIt ].

	(aBuilder shortcut: #printIt1) 
		category: #SmalltalkEditor
		default: $p command
		do: [ :morph | morph printIt ].
		
	(aBuilder shortcut: #printIt2) 
		category: #SmalltalkEditor
		default: $p ctrl
		do: [ :morph | morph printIt ].! !

!SmalltalkEditor class methodsFor: 'shortcuts' stamp: 'GuillermoPolito 5/26/2012 21:14'!
buildUnixSmalltalkEditorKeymappingsOn: aBuilder
	<keymap: #Unix>
	self buildSmalltalkEditorCtrlBasedKeymappingsOn: aBuilder! !

!SmalltalkEditor class methodsFor: 'shortcuts' stamp: 'GuillermoPolito 5/26/2012 21:15'!
buildWindowsSmalltalkEditorKeymappingsOn: aBuilder
	<keymap: #Windows>
	self buildSmalltalkEditorCtrlBasedKeymappingsOn: aBuilder! !

!SmalltalkEditor class methodsFor: 'class initialization' stamp: 'AlainPlantec 11/16/2010 21:52'!
initialize 
	"Initialize the keyboard shortcut maps and the shared buffers
	for copying text across views and managing again and undo." 
 
	"
	SmalltalkEditor initialize
	" 
	self initializeCmdKeyShortcuts.
	self initializeShiftCmdKeyShortcuts.
! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'CamilloBruni 2/26/2011 16:46'!
initializeCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."

	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"

	"SmalltalkEditor initialize"

	| cmds |
	self initializeBasicCmdKeyShortcuts.
	
	cmds := #($b #browseIt: $d #doIt: $i #inspectIt: $z #undo: $j #redo: $l #cancel: $m #implementorsOfIt: $n #sendersOfIt: $p #printIt: $q #querySymbol: $s #save: ).
	1 to: cmds size
		by: 2
		do: [:i | cmdActions at: (cmds at: i) asciiValue + 1 put: (cmds at: i + 1)].
 ! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'MarcusDenker 5/8/2012 10:16'!
initializeShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."
	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"
	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	"SmalltalkEditor initialize"

	| cmds |
 	self initializeBasicShiftCmdKeyShortcuts.
	
	cmds := #(
		$a	argAdvance:
		$d	debugIt:
		$e	methodStringsContainingIt:
		$f	displayIfFalse:
		$g	fileItIn:
		$i	exploreIt:
		$n	referencesToIt:
		$t	displayIfTrue:
		$v	pasteInitials:
		$w	methodNamesContainingIt:
	).
	1 to: cmds size by: 2 do: [ :i |
		shiftCmdActions at: ((cmds at: i) asciiValue + 1) put: (cmds at: i + 1).			"plain keys"
		shiftCmdActions at: ((cmds at: i) asciiValue - 32 + 1) put: (cmds at: i + 1).		"shifted keys"
		shiftCmdActions at: ((cmds at: i) asciiValue - 96 + 1) put: (cmds at: i + 1).		"ctrl keys"
	].! !

!SmalltalkEditor class methodsFor: 'menu declaration' stamp: 'AlainPlantec 11/16/2010 22:19'!
shiftedYellowButtonMenu
	^ (PragmaMenuBuilder 
			pragmaKeyword: self smalltalkEditorShiftedMenuKeyword
			model: nil) menu.
! !

!SmalltalkEditor class methodsFor: 'menu declaration' stamp: 'AlainPlantec 11/16/2010 17:01'!
smalltalkEditorMenuKeyword
	^'smalltalkEditorMenu'
! !

!SmalltalkEditor class methodsFor: 'menu declaration' stamp: 'EstebanLorenzano 4/26/2012 15:04'!
smalltalkEditorMenuOn: aBuilder 
	"Specify the menu used when writing code. Try it with:
	(PragmaMenuBuilder 
		pragmaKeyword: 'smalltalkEditorMenu'
		model: nil) menu popUpInWorld
	"
	<smalltalkEditorMenu>
	| uitheme|
	uitheme := UITheme current.
	(aBuilder item: #'Do it (d)' translated) selector: #doIt; help: nil; icon: uitheme  smallDoItIcon.
	(aBuilder item: #'Print it (p)' translated) selector: #printIt; help: nil; icon: uitheme  smallPrintItIcon.
	(aBuilder item: #'Inspect it (i)' translated) selector: #inspectIt; help: nil; icon: uitheme  smallInspectItIcon.
	(aBuilder item: #'Explore it (I)' translated) selector: #exploreIt; help: nil; icon: uitheme  smallInspectItIcon.
	(aBuilder item: #'Debug it (D)' translated) selector: #debugIt; help: nil; icon: uitheme  smallDebugIcon.
	(aBuilder item: #'Profile it' translated) selector: #tallyIt; icon: uitheme smallDebugIcon; withSeparatorAfter.
	(aBuilder item: #'Find...(f)' translated) selector: #find; help: nil; icon: uitheme  smallFindIcon.
	(aBuilder item: #'Find again (g)' translated) selector: #findAgain; help: nil; icon: uitheme  smallFindIcon.
	(aBuilder item: #'Extended search...' translated) selector: #shiftedTextPaneMenuRequest; withSeparatorAfter.
	(aBuilder item: #'Do again (j)' translated) selector: #again; help: nil; icon: uitheme  smallRedoIcon.
	(aBuilder item: #'Undo (z)' translated) selector: #undo; help: nil; icon: uitheme  smallUndoIcon; withSeparatorAfter.
	(aBuilder item: #'Copy (c)' translated) selector: #copySelection; help: nil; icon: uitheme  smallCopyIcon.
	(aBuilder item: #'Cut (x)' translated) selector: #cut; help: nil; icon: uitheme  smallCutIcon.
	(aBuilder item: #'Paste (v)' translated) selector: #paste; help: nil; icon: uitheme  smallPasteIcon.
	(aBuilder item: #'Paste...' translated) selector: #pasteRecent; help: nil; icon: uitheme  smallCopyIcon; withSeparatorAfter.
	(aBuilder item: #'Accept (s)' translated) selector: #accept; help: nil; icon: uitheme  smallOkIcon.
	(aBuilder item: #'Cancel (l)' translated) selector: #cancel; help: nil; icon: uitheme  smallCancelIcon.
! !

!SmalltalkEditor class methodsFor: 'menu declaration' stamp: 'AlainPlantec 11/16/2010 17:00'!
smalltalkEditorShiftedMenuKeyword
	^ 'smalltalkEditorShiftedMenu'
! !

!SmalltalkEditor class methodsFor: 'menu declaration' stamp: 'StephaneDucasse 5/13/2012 21:23'!
smalltalkEditorShiftedMenuOn: aBuilder 
	"Specify the menu used when writing code. Try it with:
	(PragmaMenuBuilder 
		pragmaKeyword: 'smalltalkEditorShiftedMenu'
		model: nil) menu popUpInWorld"

	<smalltalkEditorShiftedMenu>
	(aBuilder item: #'browse it (b)' translated) selector: #browseIt.
	(aBuilder item: #'senders of it (n)' translated) selector: #sendersOfIt.
	(aBuilder item: #'implementors of it (m)' translated) selector: #implementorsOfIt.
	(aBuilder item: #'references to it (N)' translated) selector: #referencesToIt; withSeparatorAfter.
	(aBuilder item: #'selectors containing it (W)' translated) selector: #methodNamesContainingIt.
	(aBuilder item: #'case insensitive method literal strings with it (E)' translated) selector: #methodStringsContainingit.
	(aBuilder item: #'case sensitive method literal strings with it (E)' translated) selector: #methodCaseSensitiveStringsContainingit.
	(aBuilder item: #'method source with it' translated) selector: #methodSourceContainingIt; withSeparatorAfter.
	(aBuilder item: #'class names containing it' translated) selector: #classNamesContainingIt.
	(aBuilder item: #'class comments with it' translated) selector: #classCommentsContainingIt.
	(aBuilder item: #'change sets with it' translated) selector: #browseChangeSetsWithSelector.
! !

!SmalltalkEditor class methodsFor: 'menu declaration' stamp: 'AlainPlantec 11/16/2010 22:23'!
yellowButtonMenu
	^ (PragmaMenuBuilder 
			pragmaKeyword: self smalltalkEditorMenuKeyword
			model: nil) menu.
! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
argAdvance: aKeyboardEvent
	"Invoked by Ctrl-a.  Useful after Ctrl-q.
	 Search forward from the end of the selection for a colon followed by
		a space.  Place the caret after the space.  If none are found, place the
		caret at the end of the text.  Does not affect the undoability of the 
	 	previous command."

	| start |
	self closeTypeIn.
	start := self text findString: ': ' startingAt: self stopIndex.
	start = 0 ifTrue: [ start := self text size + 1].
	self selectAt: start + 2.
	^true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'IgorStasenko 4/26/2011 16:01'!
browseChangeSetsWithSelector
	"Determine which, if any, change sets have at least one change for the selected selector, independent of class"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	Smalltalk tools changeSorter browseChangeSetsWithSelector: aSelector! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:35'!
browseClassFrom: aClassString
	"Launch a hierarchy browser for the class indicated by the current selection.  If multiple classes matching the selection exist, let the user choose among them."
	
	| selectedClass |
	self lineSelectAndEmptyCheck: [^ self].
	selectedClass := model systemNavigation 
				classFromPattern: aClassString
				withCaption: 'choose a class to browse...'.
				
	selectedClass isNil ifTrue: [^ self internalCallToImplementorsOf: aClassString ].
	
	model systemNavigation browseClass: selectedClass .! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:35'!
browseClassFromIt
	"Launch a hierarchy browser for the class indicated by the current selection.  If multiple classes matching the selection exist, let the user choose among them."
	
	self lineSelectAndEmptyCheck: [^ self].
	self browseClassFrom: (self selection string copyWithoutAll: CharacterSet crlf)! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
browseIt
	"Launch a browser for the current selection, if appropriate"
	^ self browseClassFromIt
! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
browseIt: aKeyboardEvent
	"Triggered by Cmd-B; browse the thing represented by the current selection, if plausible.  1/18/96 sw"

	self browseIt.
	^ true! !

!SmalltalkEditor methodsFor: 'accessing' stamp: 'StephaneDucasse 2/10/2011 15:03'!
caret
	"Return the index position of the caret"
	^ self startBlock stringIndex! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 5/13/2012 21:32'!
classCommentsContainingIt
	"Open a browser class comments which contain the current selection somewhere in them."

	self lineSelectAndEmptyCheck: [^ self].
	model systemNavigation  browseClassCommentsWithString: self selection string matchCase: false. ! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 7/22/2011 16:32'!
classNamesContainingIt
	"Open a browser on classes whose names contain the selected string"

	self lineSelectAndEmptyCheck: [^self].
	model systemNavigation 
		browseClassesWithNamesContaining: self selection string
		caseSensitive: Sensor shiftPressed! !

!SmalltalkEditor methodsFor: 'private' stamp: 'AlainPlantec 10/20/2011 00:29'!
codeCompletionAround:  aBlock textMorph:  aTextMorph keyStroke: anEvent
	"I'm a editor for Smalltalk, so, ok do completion around"
	Smalltalk tools codeCompletionAround: aBlock textMorph: aTextMorph keyStroke: anEvent! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'MarcusDenker 4/29/2012 13:43'!
compileSelectionFor: anObject in: evalContext
	| methodNode method failed |
	 failed :=  false.
	methodNode :=  
		[
			self class evaluatorClass new
			compileNoPattern: self selectionAsStream
			in: anObject class
			context: evalContext
			notifying: self morph
			ifFail: [ failed :=  true ] ]
		on: OutOfScopeNotification
		do: [:ex | ex resume: true].
	failed ifTrue: [ ^ nil ].	
	^methodNode generateWithSource.
	! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'MarcusDenker 10/21/2011 17:17'!
debug: aCompiledMethod receiver: anObject in: evalContext

 	 | guineaPig debugger context |

       guineaPig := [
              aCompiledMethod
                       valueWithReceiver: anObject
                       arguments: (evalContext ifNil: [ #() ] ifNotNil: [ { evalContext } ]) ] newProcess.
       context := guineaPig suspendedContext.
       debugger :=  Smalltalk tools debugger new
               process: guineaPig
               controller: nil
               context: context.
       debugger openFullNoSuspendLabel: 'Debug it'.
       [ debugger interruptedContext method == aCompiledMethod ]
               whileFalse: [ debugger send ]! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'StephaneDucasse 3/1/2011 19:38'!
debugIt

	| method receiver context |
	(model respondsTo: #doItReceiver) 
		ifTrue: 
			[receiver := model doItReceiver.
			context := model doItContext]
		ifFalse:
			[receiver := context := nil].
	self lineSelectAndEmptyCheck: [^self].
	method := self compileSelectionFor: receiver in: context.
	method notNil ifTrue:
		[self debug: method receiver: receiver in: context].! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'CamilloBruni 9/30/2011 10:27'!
debugIt: aKeyboardEvent
	"Evaluate the current selection in the Debugger."

	self debugIt.
	^ true! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
displayIfFalse: aKeyboardEvent
	"Replace the current text selection with the text 'ifFalse:'--initiated by 
	ctrl-f."

	self addString: 'ifFalse:'.
	^false! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
displayIfTrue: aKeyboardEvent
	"Replace the current text selection with the text 'ifTrue:'--initiated by 
	ctrl-t."

	self addString: 'ifTrue:'.
	^false! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'AlainPlantec 11/8/2010 22:09'!
doIt
	"Set the context to include pool vars of the model.  Then evaluate."
	^ self evaluateSelection! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
doIt: aKeyboardEvent
	"Called when user hits cmd-d.  Select the current line, if relevant, then evaluate and execute.  2/1/96 sw.
	2/29/96 sw: don't call selectLine; it's done by doIt now"

	self doIt.
	^ true! !

!SmalltalkEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/9/2010 15:05'!
editingStateClass
	^ SmalltalkEditingState! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'nice 3/31/2011 22:38'!
evaluateSelection
	"Treat the current selection as an expression; evaluate it and return the result"
	
	^self evaluateSelectionAndDo: [:result | result]! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'nice 3/31/2011 22:38'!
evaluateSelectionAndDo: aBlock
	"Treat the current selection as an expression; evaluate it and invoke aBlock with the result."
	| result rcvr ctxt |

	self lineSelectAndEmptyCheck: [^ ''].
	(model respondsTo: #doItReceiver) 
		ifTrue: [ rcvr := model doItReceiver.
				ctxt := model doItContext]
		ifFalse: [ rcvr := ctxt := nil].
	result := [
		rcvr class evaluatorClass new 
			evaluate: self selectionForDoitAsStream
			in: ctxt
			to: rcvr
			notifying: self morph  "don't set it to self.. The receiver can be destroyed as a result of evaluateSelection"
			ifFail: [self morph flash. ^nil]
			logged: true.
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
	^ aBlock value: result! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'nice 3/31/2011 22:45'!
exploreIt
	self evaluateSelectionAndDo: [:result | result explore]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
exploreIt: aKeyboardEvent
	"Explore the selection -- invoked via cmd-shift-I.  If there is no current selection, use the current line."

	self exploreIt.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
fileItIn
	"Make a Stream on the text selection and fileIn it.
	 1/24/96 sw: moved here from FileController; this function can be useful from any text window that shows stuff in chunk format"

	| selection |
	selection := self selection.
	(ReadWriteStream on: selection string from: 1 to: selection size) fileIn
! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
fileItIn: aKeyboardEvent
	"File in the selection; invoked via a keyboard shortcut, -- for now, cmd-shift-G."

	self fileItIn.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:37'!
implementorsOf: aSelector
	"Open an implementors browser on the given selector"

	(model systemNavigation allImplementorsOf: aSelector) isEmpty
			ifTrue: [self internalCallToBrowse: aSelector ]
			ifFalse: [ model systemNavigation browseAllImplementorsOf: aSelector]! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:38'!
implementorsOfIt
	"Open an implementors browser on the selected selector"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	self implementorsOf: aSelector! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
implementorsOfIt: aKeyboardEvent
	"Triggered by Cmd-m; browse implementors of the selector represented by the current selection, if plausible. 2/1/96 sw"

	self implementorsOfIt.
	^ true! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'nice 3/31/2011 22:45'!
inspectIt
	self evaluateSelectionAndDo: [:result | result inspect]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
inspectIt: aKeyboardEvent
	"Inspect the selection -- invoked via cmd-i.  If there is no current selection, use the current line.  1/17/96 sw
	 2/29/96 sw: don't call selectLine; it's done by inspectIt now"

	self inspectIt.
	^ true! !

!SmalltalkEditor methodsFor: 'shortcuts' stamp: 'GuillermoPolito 5/21/2012 01:31'!
installKeymappingsOn: aMorph
	super installKeymappingsOn: aMorph.
	aMorph attachKeymapCategory: #SmalltalkEditor! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:39'!
internalCallToBrowse: aSelector
	"Launch a browser for the class indicated by the given selector. 
	If multiple classes matching the selection exist, let the user choose among them."
	| aClass |
	aClass := model systemNavigation 
				classFromPattern: aSelector
				withCaption: 'choose a class to browse...'.
	aClass ifNil: [^ self].
	model systemNavigation browseClass: aClass .! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:39'!
internalCallToBrowseIt
	"Launch a browser for the class indicated by the current selection. 
	If multiple classes matching the selection exist, let the user choose among them."
	| aClass |
	self lineSelectAndEmptyCheck: [^ self].
	self internalCallToBrowse: (self selection string copyWithoutAll: CharacterSet crlf)! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:36'!
internalCallToImplementorsOf: aSelector
	"Open an implementors browser on the given selector"
	
	 (model systemNavigation allImplementorsOf: aSelector) isEmpty
				ifFalse: [ model systemNavigation browseAllImplementorsOf: aSelector]
			
	! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:36'!
internalCallToImplementorsOfIt
	"Open an implementors browser on the selected selector"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	self internalCallToImplementorsOf: aSelector.
			
	! !

!SmalltalkEditor methodsFor: 'testing' stamp: 'StephaneDucasse 2/23/2012 13:04'!
isCaretBehindChar
	"Return true if the cursor position is after an alphanumeric character, otherwise false."

	| cursorPosition |
	^(cursorPosition := self startIndex) >= 2 and: [
		(self text at: cursorPosition - 1) isAlphaNumeric ]! !

!SmalltalkEditor methodsFor: 'testing' stamp: 'AlainPlantec 11/8/2010 22:09'!
isSmalltalkEditor
	^ true ! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 5/13/2012 21:25'!
methodCaseSensitiveStringsContainingit
	"Open a browser on methods which contain the current selection as part of a string constant."

	self lineSelectAndEmptyCheck: [^ self].
	
	model systemNavigation  browseMethodsWithString: self selection string matchCase: true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'SvenVanCaekenberghe 1/8/2012 14:48'!
methodNamesContainingIt
	"Open a browser on methods names containing the selected string"

	self lineSelectAndEmptyCheck: [ ^ self ].
	Cursor wait showWhile: [ 
		model systemNavigation  browseMethodsWhoseNamesContain: self selection string trimBoth ]
	! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
methodNamesContainingIt: aKeyboardEvent
	"Browse methods whose selectors containing the selection in their names"

	self methodNamesContainingIt.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 5/13/2012 21:33'!
methodSourceContainingIt
	"Open a browser on methods which contain the current selection in their source (case-sensitive full-text search of source).   EXTREMELY slow!!"
	self lineSelectAndEmptyCheck: [^ self].
	model systemNavigation  browseMethodsWithSourceString: self selection string matchCase: false.! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
methodStringsContainingIt: aKeyboardEvent
	"Invoked from cmd-E -- open a browser on all methods holding string constants containing it."

	self methodStringsContainingit.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 5/13/2012 21:27'!
methodStringsContainingit
	"Open a browser on methods which contain the current selection as part of a string constant."

	self lineSelectAndEmptyCheck: [^ self].
	model systemNavigation  browseMethodsWithString: self selection string matchCase: false.
! !

!SmalltalkEditor methodsFor: 'private' stamp: 'BenjaminVanRyseghem 4/14/2011 14:07'!
modelCurrentSelectedClass
	" ugly dispatch, but current Browser protocol names aren't really cool "
	" for Nautilus ( and other tools), selectedClass sounds better, but I think it shouldn't be hardcoded "
	" IT'S A TEMP SOLUTION -- BenjaminVanRyseghem 14/04/11"

	" From now, all tools should use #selectedBehavior, waiting for a better solution - Thx Nice "
	(model respondsTo: #selectedBehavior)
		ifTrue: [^ model selectedBehavior ].
	
	" old selectors in order of uses "
	(model respondsTo: #selectedClass)
		ifTrue: [^ model selectedClass ].
	(model respondsTo: #selectedClassOrMetaClass)
		ifTrue: [ ^ model selectedClassOrMetaClass ].
	
		
	^ nil! !

!SmalltalkEditor methodsFor: 'new selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
nextTokenFrom: start direction: dir
	"simple token-finder for compiler automated corrections"
	| loc str |
	loc := start + dir.
	str := self string.
	[(loc between: 1 and: str size) and: [(str at: loc) isSeparator]]
		whileTrue: [loc := loc + dir].
	^ loc! !

!SmalltalkEditor methodsFor: 'new selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
notify: aString at: anInteger in: aStream 
	"The compilation of text failed. The syntax error is noted as the argument, 
	aString. Insert it in the text at starting character position anInteger."

	| pos |
	pos := self selectionInterval notEmpty
		ifTrue: [self startIndex + anInteger - 1 ]
		ifFalse: [anInteger].
	self insertAndSelect: aString at: (pos max: 1)! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
pasteInitials: aKeyboardEvent 
	"Replace the current text selection by an authorship name/date stamp; invoked by cmd-shift-v, easy way to put an authorship stamp in the comments of an editor."

	self closeTypeIn.
	self replace: self selectionInterval with: (Text fromString: Author changeStamp) and: [self selectAt: self stopIndex].
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'BenjaminVanRyseghem 4/14/2011 13:31'!
prettyPrint
	"Reformat the contents of the receiver's view (a Browser)."

	| selectedClass newText |
	model selectedMessageName ifNil: [^ morph flash].
	selectedClass := self modelCurrentSelectedClass.
	newText := selectedClass compilerClass new
		format: self text
		in: selectedClass
		notifying: self.
	newText ifNotNil:
		[ self selectInvisiblyFrom: 1 to: self paragraph text size.
		self replaceSelectionWith: (newText asText makeSelectorBoldIn: selectedClass).
		self selectAt: 1]! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'nice 3/31/2011 22:45'!
printIt
	"Treat the current text selection as an expression; evaluate it. Insert the 
	description of the result of evaluation after the selection and then make 
	this description the new text selection."
	self evaluateSelectionAndDo: [:result | self afterSelectionInsertAndSelect: result printString]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
printIt: aKeyboardEvent
	"Print the results of evaluting the selection -- invoked via cmd-p.  If there is no current selection, use the current line.  1/17/96 sw
	 2/29/96 sw: don't call selectLine now, since it's called by doIt"

	self printIt.
	^ true! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/14/2010 22:38'!
querySymbol: aKeyboardEvent
	"Invoked by Ctrl-q to query the Symbol table and display alternate symbols.
	 See comment in completeSymbol:lastOffering: for details."

	self closeTypeIn.
	self editingState closeSelectorChooser.
	self editingState querySymbolFrom: morph.
	^true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:41'!
referencesTo: aVariableOrClassName
	"Open a references browser on the given symbol"
	| env ref |

	" ugly dispatch, but current Browser protocol names aren't really cool "
	env := self modelCurrentSelectedClass ifNil: [ Smalltalk globals ].

	env isBehavior ifTrue: [
		(env instVarIndexFor: aVariableOrClassName) > 0 ifTrue: [
		 ^ self systemNavigation browseAllAccessesTo: aVariableOrClassName from: env]].

	ref:= (env bindingOf: aVariableOrClassName) ifNil: [ ^ self ].
	
	self systemNavigation browseAllCallsOn: ref! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:40'!
referencesToIt
	"Open a references browser on the selected symbol"

	| aSymbol env ref |
	self selectLine.
	((aSymbol := self selectedSymbol) == nil)
			ifTrue: [^ morph flash].

	self referencesTo: aSymbol! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
referencesToIt: aKeyboardEvent
	"Triggered by Cmd-N; browse references to the current selection"

	self referencesToIt.
	^ true! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
save: aKeyboardEvent
	"Submit the current text.  Equivalent to 'accept' 1/18/96 sw"

	self closeTypeIn.
	self accept.
	^ true! !

!SmalltalkEditor methodsFor: 'new selection' stamp: 'AlainPlantec 2/21/2011 20:02'!
selectLine
	"Make the receiver's selection, if it currently consists of an insertion point only, encompass the current line."
	| i left right s |
	self hasSelection ifTrue: [^self].
	s := self string.
	i := s encompassParagraph: self selectionInterval.
	left := i first.
	right := i last.
	[ left <= right and: [ (s at: left) = $" or: [(s at: left) isSeparator] ] ]
		whileTrue: [ left := left + 1 ].
	[ left <= right and: [ (s at: right) = $"  or: [(s at: right) isSeparator] ] ]
		whileTrue: [ right := right - 1 ].
	self selectInterval: (left to: right)! !

!SmalltalkEditor methodsFor: 'new selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
selectPrecedingIdentifier
	"Invisibly select the identifier that ends at the end of the selection, if any."

	| string sep stop tok |
	tok := false.
	string := self string.
	stop := self stopIndex - 1.
	[stop > 0 and: [(string at: stop) isSeparator]] whileTrue: [stop := stop - 1].
	sep := stop.
	[sep > 0 and: [(string at: sep) tokenish]] whileTrue: [tok := true. sep := sep - 1].
	tok ifTrue: [self selectInvisiblyFrom: sep + 1 to: stop]! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
selectedSelector
	"Try to make a selector out of the current text selection"
	^self selection string findSelector! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
selectedSymbol
	"Return the currently selected symbol, or nil if none.  Spaces, tabs and returns are ignored"

	| aString |
	self hasCaret ifTrue: [^ nil].
	aString := self selection string copyWithoutAll:
		{Character space.  Character cr.  Character tab}.
	aString size = 0 ifTrue: [^ nil].
	Symbol hasInterned: aString  ifTrue: [:sym | ^ sym].

	^ nil! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'AlainPlantec 11/8/2010 22:09'!
selectionForDoitAsStream
	"Answer a ReadStream on the text in the paragraph that is currently
	selected. "
	^ ReadWriteStream
		on: self string
		from: self startIndex 
		to: self stopIndex - 1! !

!SmalltalkEditor methodsFor: 'accessing' stamp: 'StephaneDucasse 2/23/2012 13:04'!
selectionPosition: aString

	| startIndex |
	startIndex := self startIndex.
	^[
		| bottomLeft topLeft index |
		index := startIndex - aString size.
		self selectInvisiblyFrom: index to: index - 1.
		bottomLeft := self paragraph selectionRects first bottomLeft.
		topLeft := self morph owner submorphBounds topLeft.
		topLeft + bottomLeft ]
			ensure: [ self selectInvisiblyFrom: startIndex to: startIndex - 1 ]! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'SvenVanCaekenberghe 1/8/2012 14:48'!
selectionString
	"remove the initial and final quote marks, if present"
	"'''h''' withoutWrappingDoubleQuotes"
	| quote selection  |
	selection :=  self selection string trimBoth .
	selection size < 2 ifTrue: [ ^ selection ].
	quote := selection first.
	^ quote = $"
		ifTrue: [ selection copyFrom: 2 to: selection size - 1 ]
		ifFalse: [ selection ].! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:31'!
sendersOf: selectedSelector
	" opens a Browser on the given selector"
	self model systemNavigation browseAllCallsOn: selectedSelector! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'CamilloBruni 2/1/2012 19:30'!
sendersOfIt
	"Open a senders browser on the selected selector"

	| selectedSelector |
	self lineSelectAndEmptyCheck: [^ self].
	selectedSelector  := self selectedSelector.
	selectedSelector isNil ifTrue: [^ morph flash].
	self sendersOf: selectedSelector! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
sendersOfIt: aKeyboardEvent
	"Triggered by Cmd-n; browse implementors of the selector represented by the current selection, if plausible. 2/1/96 sw"

	self sendersOfIt.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
spawn
	"Create and schedule a message browser for the code of the model's 
	selected message. Retain any edits that have not yet been accepted."
	| code |
	code := self string.
	self cancel.
	model spawn: code.! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
spawnIt: aKeyboardEvent
	"Triggered by Cmd-o; spawn a new code window, if it makes sense."

	self spawn.
	^ true! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'AlainPlantec 11/8/2010 22:09'!
tallyIt

	^ self tallySelection! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'AlainPlantec 7/22/2011 20:28'!
tallySelection
	"Treat the current selection as an expression; evaluate and tally it."
	| receiver context v |
	(model respondsTo: #doItReceiver) 
		ifTrue: 
			[receiver := model doItReceiver.
			context := model doItContext]
		ifFalse:
			[receiver := context := nil].
	self lineSelectAndEmptyCheck: [ ^ self ].

	[ | compiledMethod profiler |
		compiledMethod := self compileSelectionFor: receiver in: context.
		compiledMethod ifNil: [^ self].
		profiler := TimeProfiler spyOn: [ v:= compiledMethod valueWithReceiver: receiver arguments: #()].
		profiler showResult: v.
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
! !

!SmalltalkEditor methodsFor: 'accessing' stamp: 'StephaneDucasse 2/23/2012 13:04'!
wordAtCaret

	^self paragraph text asString wordBefore: self startIndex - 1
! !

!TextEditor class methodsFor: 'shortcuts' stamp: 'GuillermoPolito 5/26/2012 21:05'!
buildTextEditorCtrlBasedKeymapsOn: aBuilder
		
	(aBuilder shortcut: #selectAll) 
		category: #TextEditor
		default: $a ctrl
		do: [ :target | target handleEdit: [ target selectAll ] ].
		
	(aBuilder shortcut: #copySelection) 
		category: #TextEditor
		default: $c ctrl
		do: [ :target | target copySelection ].
		
	(aBuilder shortcut: #find) 
		category: #TextEditor
		default: $f ctrl
		do: [ :target | target find ].
		
	(aBuilder shortcut: #findAgain) 
		category: #TextEditor
		default: $g ctrl
		do: [ :target | target findAgain ].
		
	(aBuilder shortcut: #setSearchString)
		category: #TextEditor
		default: $h ctrl
		do: [ :target | target setSearchString ].
		
	(aBuilder shortcut: #cancel) 
		category: #TextEditor
		default: $l ctrl
		do: [ :target | target cancel ].
		
	(aBuilder shortcut: #paste) 
		category: #TextEditor
		default: $v ctrl
		do: [ :target | target paste ].
		
	(aBuilder shortcut: #cut) 
		category: #TextEditor
		default: $x ctrl
		do: [ :target | target cut ].
				
	(aBuilder shortcut: #undo) 
		category: #TextEditor
		default: $z ctrl
		do: [ :target | target undo ].
		
	(aBuilder shortcut: #redo) 
		category: #TextEditor
		default: $z ctrl shift
		do: [ :target | target redo ].
	
	(aBuilder shortcut: #cursorHome)
		category: #TextEditor
		default: Character home ctrl
		do: [ :target :morph :event | target cursorHome: event ].
	
	(aBuilder shortcut: #cursorEnd)
		category: #TextEditor
		default: Character end ctrl
		do: [ :target :morph :event | target cursorEnd: event ].! !

!TextEditor class methodsFor: 'shortcuts' stamp: 'GuillermoPolito 5/26/2012 21:06'!
buildUnixTextEditorKeymapsOn: aBuilder
	<keymap: #Unix>
	self buildTextEditorCtrlBasedKeymapsOn: aBuilder! !

!TextEditor class methodsFor: 'shortcuts' stamp: 'GuillermoPolito 5/26/2012 21:13'!
buildWindowsTextEditorKeymapsOn: aBuilder
	<keymap: #Windows>
	self buildTextEditorCtrlBasedKeymapsOn: aBuilder! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/8/2010 22:09'!
caseSensitiveFinds
	^ CaseSensitiveFinds ifNil: [CaseSensitiveFinds := false]! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/8/2010 22:09'!
caseSensitiveFinds: aBoolean
	CaseSensitiveFinds := aBoolean! !

!TextEditor class methodsFor: 'accessing' stamp: 'CamilloBruni 2/26/2011 15:28'!
cmdActions
	^cmdActions! !

!TextEditor class methodsFor: 'instance creation' stamp: 'IgorStasenko 12/12/2010 03:36'!
forMorph: aMorph

	^ self basicNew initWithMorph: aMorph! !

!TextEditor class methodsFor: 'class initialization' stamp: 'AlainPlantec 11/16/2010 22:22'!
initialize 
	"Initialize the keyboard shortcut maps and the shared buffers
	for copying text across views and managing again and undo." 
 
	"
	TextEditor initialize
	" 

	self initializeCmdKeyShortcuts.
	self initializeShiftCmdKeyShortcuts.
! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'CamilloBruni 2/6/2012 11:38'!
initializeBasicCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table.
	
	TextEditor initialize
	"
	| cmdMap |
	
	cmdMap := Array new: 256 withAll: #noop:.		"use temp in case of a crash"
	cmdMap at: 1 + 1 put: #cursorHome:.				"home key"
	cmdMap at: 4 + 1 put: #cursorEnd:.				"end key"
	cmdMap at: 8 + 1 put: #backspace:.				"ctrl-H or delete key"
	cmdMap at: 11 + 1 put: #cursorPageUp:.			"page up key"
	cmdMap at: 12 + 1 put: #cursorPageDown:.		"page down key"
	cmdMap at: 13 + 1 put: #crWithIndent:.			"cmd-Return"
	cmdMap at: 27 + 1 put: #escape:.		"escape key"
	cmdMap at: 28 + 1 put: #cursorLeft:.				"left arrow key"
	cmdMap at: 29 + 1 put: #cursorRight:.			"right arrow key"
	cmdMap at: 30 + 1 put: #cursorUp:.				"up arrow key"
	cmdMap at: 31 + 1 put: #cursorDown:.			"down arrow key"
	cmdMap at: 32 + 1 put: #selectWord:.			"space bar key"
	cmdMap at: 127 + 1 put: #forwardDelete:.		"del key"
	cmdMap at: 115 + 1 put:#accept:.	
	'([{''"<' do: [:char | cmdMap at: char asciiValue + 1 put: #enclose:].
	
	cmdMap at: $, asciiValue + 1 put: #shiftEnclose:.
	
	#(
		$a #selectAll: 
		$c #copySelection: 
		$e #exchange: 
		$f #find: 
		$g #findAgain: 
		$h #setSearchString: 
		$v #paste: 
		$w #backWord: 
		$x #cut: 
		$y #swapChars: 
		$z #undo: 
		$j #redo: 
		"$u #exploreUndoManager: ")
	pairsDo: [ :char :symbol|
		cmdMap at: char asciiValue + 1 put: symbol].
		
	cmdActions := cmdMap! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'CamilloBruni 2/6/2012 11:38'!
initializeBasicShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."

	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	"TextEditor initialize"
	
	| cmdMap |

	"shift-command and control shortcuts"
	cmdMap := Array new: 256 withAll: #noop:.  		"use temp in case of a crash"
	cmdMap at: ( 1 + 1) put: #cursorHome:.			"home key"
	cmdMap at: ( 4 + 1) put: #cursorEnd:.				"end key"
	cmdMap at: ( 8 + 1) put: #forwardDelete:.			"ctrl-H or delete key"
	cmdMap at: (11 + 1) put: #cursorPageUp:.			"page up key"
	cmdMap at: (12 + 1) put: #cursorPageDown:.		"page down key"
	cmdMap at: (13 + 1) put: #crWithIndent:.			"ctrl-Return"
	cmdMap at: (27 + 1) put: #offerMenuFromEsc:.	"escape key"
	cmdMap at: (28 + 1) put: #cursorLeft:.			"left arrow key"
	cmdMap at: (29 + 1) put: #cursorRight:.			"right arrow key"
	cmdMap at: (30 + 1) put: #cursorUp:.				"up arrow key"
	cmdMap at: (31 + 1) put: #cursorDown:.			"down arrow key"
	cmdMap at: (32 + 1) put: #selectWord:.			"space bar key"
	cmdMap at: (45 + 1) put: #changeEmphasis:.		"cmd-sh-minus"
	cmdMap at: (61 + 1) put: #changeEmphasis:.		"cmd-sh-plus"
	cmdMap at: (127 + 1) put: #forwardDelete:.		"del key"

	"Note: Command key overrides shift key, so, for example, cmd-shift-9 produces $9 not $("
	'9[,''' do: [ :char | cmdMap at: (char asciiValue + 1) put: #shiftEnclose: ].	"({< and double-quote"
	"Note: Must use cmd-9 or ctrl-9 to get '()' since cmd-shift-9 is a Mac FKey command."

	"NB: sw 12/9/2001 commented out the idiosyncratic line just below, which was grabbing shift-esc in the text editor and hence which argued with the wish to have shift-esc be a universal gesture for escaping the local context and calling up the desktop menu."  
	"cmdMap at: (27 + 1) put: #shiftEnclose:." 	"ctrl-["

	"'""''(' do: [ :char | cmdMap at: (char asciiValue + 1) put: #enclose:]."

	#(
		$c	compareToClipboard:
		$d	duplicate:
		$h	cursorTopHome:
		$j	doAgainMany:
		$k	changeStyle:
		$l	outdent:
		$m selectCurrentTypeIn:
		$r	indent:
		$s	search:
		$u	changeLfToCr:
		$x	makeLowercase:
		$y	makeUppercase:
		$z	makeCapitalized:
		$g #findAgain:
	) pairsDo: [ :char :symbol|
		cmdMap at: char asciiValue + 1 put: symbol.			"plain keys"
		cmdMap at: char asciiValue - 32 + 1 put: symbol.		"shifted keys"
		cmdMap at: char asciiValue - 96 + 1 put: symbol.		"ctrl keys"
	].
	shiftCmdActions := cmdMap! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'AlainPlantec 11/8/2010 22:09'!
initializeCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."



	"
	TextEditor initialize
	"

	| cmds |
	self initializeBasicCmdKeyShortcuts.
			
	'0123456789-=' 
		do: [:char | cmdActions at: char asciiValue + 1 put: #changeEmphasis:].
	
	cmds := #($k #offerFontMenu: $u #align: ).
	1 to: cmds size
		by: 2
		do: [:i | self cmdActions at: (cmds at: i) asciiValue + 1 put: (cmds at: i + 1)]! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'AlainPlantec 11/16/2010 22:00'!
initializeShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."
	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	"TextEditor initialize"
	
	| cmds |

	self initializeBasicShiftCmdKeyShortcuts.
	self shiftCmdActions at: (45 + 1) put: #changeEmphasis:.		"cmd-sh-minus"
	self shiftCmdActions at: (61 + 1) put: #changeEmphasis:.		"cmd-sh-plus"

	cmds := #(
		$k	changeStyle:
		"$u #clearUndoManager:	"
	).
	1 to: cmds size by: 2 do: [ :i |
		self shiftCmdActions at: ((cmds at: i) asciiValue + 1) put: (cmds at: i + 1).			"plain keys"
		self shiftCmdActions at: ((cmds at: i) asciiValue - 32 + 1) put: (cmds at: i + 1).		"shifted keys"
		self shiftCmdActions at: ((cmds at: i) asciiValue - 96 + 1) put: (cmds at: i + 1).		"ctrl keys"
	].! !

!TextEditor class methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
shiftCmdActions
	^shiftCmdActions! !

!TextEditor class methodsFor: 'menu declaration' stamp: 'AlainPlantec 11/16/2010 22:15'!
shiftedYellowButtonMenu
	"Answer the menu to be presented when the yellow button is pressed while the shift key is down"

	^ (PragmaMenuBuilder pragmaKeyword: 'textEditorShiftedMenu' model: nil) menu! !

!TextEditor class methodsFor: 'menu declaration' stamp: 'FernandoOlivero 4/12/2011 10:40'!
textEditorMenuOn: aBuilder 
	"Specify the menu used when writing text. Try it with:
	(PragmaMenuBuilder 
		pragmaKeyword: 'textEditorMenu'
		model: nil) menu popUpInWorld"

	<textEditorMenu>
	| uitheme |
	uitheme := UITheme current.
	(aBuilder item: #'Find...(f)' translated) selector: #find; help: nil; icon: uitheme smallFindIcon.
	(aBuilder item: #'Find again (g)' translated) selector: #findAgain; help: nil; icon: uitheme smallFindIcon.
	(aBuilder item: #'Set search string (h)' translated) selector: #setSearchString.
	(aBuilder item: #'Do again (j)' translated) selector: #again; help: nil; icon: uitheme  smallRedoIcon.
	(aBuilder item: #'Undo (z)' translated) selector: #undo; help: nil; icon: uitheme smallUndoIcon; withSeparatorAfter.
	(aBuilder item: #'Copy (c)' translated) selector: #copySelection; help: nil; icon: uitheme  smallCopyIcon.
	(aBuilder item: #'Cut (x)' translated) selector: #cut; help: nil; icon:  uitheme smallCutIcon.
	(aBuilder item: #'Paste (v)' translated) selector: #paste; help: nil; icon:  uitheme smallPasteIcon.
	(aBuilder item: #'Paste...' translated) selector: #pasteRecent; help: nil; icon: uitheme  smallCopyIcon; withSeparatorAfter.
	(aBuilder item: #'Set font... (k)' translated) selector: #offerFontMenu.
	(aBuilder item: #'Set style... (K)' translated) selector: #changeStyle.
	(aBuilder item: #'Set alignment...' translated) selector: #chooseAlignment; withSeparatorAfter.
	(aBuilder item: #'Accept (s)' translated) selector: #accept; help: nil; icon: uitheme  smallOkIcon.
	(aBuilder item: #'Cancel (l)' translated) selector: #cancel; help: nil; icon: uitheme  smallCancelIcon.

! !

!TextEditor class methodsFor: 'menu declaration' stamp: 'AlainPlantec 11/16/2010 22:03'!
textEditorShiftedMenuOn: aBuilder 
	"Specify the menu used when writing text. Try it with:
	(PragmaMenuBuilder 
		pragmaKeyword: 'textEditorShiftedMenu'
		model: nil) menu popUpInWorld"

	<textEditorShiftedMenu>
	(aBuilder item: #'set font... (k)' translated) selector: #offerFontMenu.
	(aBuilder item: #'set style... (K)' translated) selector: #changeStyle.
	(aBuilder item: #'set alignment...' translated) selector: #chooseAlignment.
	(aBuilder item: #'more...' translated) selector: #yellowButtonActivity.
! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/12/2010 22:35'!
useFindReplaceSelection
	^ UseFindReplaceSelection ifNil: [UseFindReplaceSelection := true]! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/13/2010 00:12'!
useFindReplaceSelection: aBoolean
	UseFindReplaceSelection := aBoolean! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/12/2010 11:50'!
useSecondarySelection
	^ UseSecondarySelection ifNil: [UseSecondarySelection := true]! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/12/2010 11:53'!
useSecondarySelection: aBoolean
	UseSecondarySelection := aBoolean! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/17/2010 15:25'!
useSelectionBar
	^ UseSelectionBar ifNil: [UseSelectionBar := true]! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/17/2010 15:25'!
useSelectionBar: aBoolean
	UseSelectionBar := aBoolean! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/8/2010 22:09'!
walkAlongDisplayedLine
	^ WalkAlongDisplayedLine ifNil: [WalkAlongDisplayedLine := true]! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 11/8/2010 22:09'!
walkAlongDisplayedLine: aBoolean
	WalkAlongDisplayedLine := aBoolean! !

!TextEditor class methodsFor: 'menu declaration' stamp: 'AlainPlantec 11/16/2010 22:21'!
yellowButtonMenu
	^ (PragmaMenuBuilder pragmaKeyword: 'textEditorMenu' model: nil) menu! !

!TextEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 3/25/2011 11:27'!
accept
	"Save the current text of the text being edited as the current acceptable version for purposes of canceling.  Allow my morph to take appropriate action"
	morph canChangeText 
		ifTrue: [morph acceptContents]! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 12/30/2010 09:26'!
accept: aKeyboardEvent 
	"Save the current text of the text being edited as the current acceptable version for purposes of canceling.  Allow my morph to take appropriate action"
	self accept.
	^ true! !

!TextEditor methodsFor: 'typing support' stamp: 'CamilloBruni 2/26/2011 14:23'!
acceptCR: aKeyboardEvent return: return
	((aKeyboardEvent keyValue = Character cr asciiValue) and: [morph acceptOnCR])
		ifTrue: [
			self closeTypeIn.
			return value: true].! !

!TextEditor methodsFor: 'undo-redo private' stamp: 'AlainPlantec 12/14/2010 14:43'!
addDeleteSelectionUndoRecord
	| undoText redoText undoRec |
	undoText := self selection.
	redoText := self nullText.
	self editingState
		redoArray: { morph. #undoTypeIn:interval:.  {redoText. self selectionInterval} } 
		undoArray: {morph. #redoTypeIn:interval:.  {undoText. self selectionInterval first to: self selectionInterval first - 1} }.
! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 11/8/2010 22:09'!
addString: aString
	"Think of a better name"
	self zapSelectionWith: (Text string: aString attributes: self emphasisHere)! !

!TextEditor methodsFor: 'undo-redo private' stamp: 'AlainPlantec 12/14/2010 14:44'!
addTypeInUndoRecord
	| begin stop undoText redoText undoRec |
	begin := self startOfTyping min: self stopIndex. 
	stop := self stopIndex max: self startOfTyping.
	self editingState 
		previousInterval: (begin to: stop - 1) 
		selection: (self text copyFrom: begin to: stop - 1). 
	undoText := self nullText.
	redoText := stop > begin
			ifTrue: [self text copyFrom: begin to: stop - 1]
			ifFalse: [self nullText].
	((undoText isEmpty and: [redoText isEmpty]) and: [self editingState previousInterval size < 1])
			ifFalse: [self editingState
				redoArray: { morph. #redoTypeIn:interval:.  {redoText. begin to: begin + self selection size - 1} } 
				undoArray: {morph. #undoTypeIn:interval:.  {undoText. begin to: stop - 1} }].
	
! !

!TextEditor methodsFor: 'new selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
afterSelectionInsertAndSelect: aString

	self insertAndSelect: aString at: self stopIndex ! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 2/14/2011 21:20'!
again
	"Text substitution. If the left shift key is down, the substitution is
	made throughout the entire Paragraph. Otherwise, only the next 
	possible substitution is made"
	true.
	! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
align
	"Align text according to the next greater alignment value,
	cycling among leftFlush, rightFlush, center, and justified."
	self changeAlignment.
	self recomputeInterval! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
align: aKeyboardEvent
	"Triggered by Cmd-u;  cycle through alignment alternatives.  8/11/96 sw"

	self align.
	^ true! !

!TextEditor methodsFor: 'constants' stamp: 'StephaneDucasse 5/25/2011 17:36'!
alignmentChoices
	"Return the symbols representing the TextAlignment operations to get a text aligned."
	
	^ #(leftFlush centered justified rightFlush)! !

!TextEditor methodsFor: 'private' stamp: 'Sd 1/28/2011 14:42'!
applyAttribute: aTextAttribute
	"The user selected aTextAttribute via shortcut, menu or other means.
	If there is a selection, apply the attribute to the selection.
	In any case use the attribute for the user input (emphasisHere)"

	self editingState emphasisHere: (Text addAttribute: aTextAttribute toArray: self emphasisHere).
	self selection
		ifEmpty: [ self paragraph composeAll ]
		ifNotEmpty: [:sel | self replaceSelectionWith: (sel addAttribute: aTextAttribute).
			self paragraph composeAll.
			self recomputeSelection]! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 12/14/2010 22:49'!
backTo: startIndex
	"Backspace typing"
	self stopIndex > startIndex ifTrue: [
		self selectFrom: startIndex to: self stopIndex - 1.
		self addDeleteSelectionUndoRecord.
		self zapSelectionWith: self nullText].
	^ false


	! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 12/14/2010 22:49'!
backspace: aKeyboardEvent
	| result |
	self closeTypeIn.
	result := super backspace: aKeyboardEvent.
	self editingState keystroke: aKeyboardEvent from: morph.
	^ result! !

!TextEditor methodsFor: 'binding' stamp: 'AlainPlantec 11/8/2010 22:09'!
bindingOf: aString
	^model bindingOf: aString! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 9/15/2011 17:26'!
blankParagraph
	| blankParagraph |
	blankParagraph  := Paragraph new.
	blankParagraph 
		compose:  '' asText 
		style:  TextStyle default copy    
		from: 1
		in:  (0@0 extent: 10@10) .
	^ blankParagraph	! !

!TextEditor methodsFor: 'parenblinking' stamp: 'AlainPlantec 11/16/2010 10:15'!
blinkParen
	| lastParenLocation |
	(lastParenLocation := self editingState lastParenLocation) ifNotNil: [
		self text string size >= lastParenLocation ifTrue: [
			self text
				addAttribute: TextEmphasis bold
				from: lastParenLocation
				to: lastParenLocation]]! !

!TextEditor methodsFor: 'parenblinking' stamp: 'AlainPlantec 11/8/2010 22:09'!
blinkParenAt: parenLocation 
	self text
		addAttribute: TextEmphasis bold
		from: parenLocation
		to: parenLocation.
	self editingState lastParenLocation: parenLocation.! !

!TextEditor methodsFor: 'parenblinking' stamp: 'AlainPlantec 11/8/2010 22:09'!
blinkPrevParen: aCharacter
	| openDelimiter closeDelimiter level string here hereChar |
	string := self string.
	here := self editingState pointBlock stringIndex.
	openDelimiter := "aKeyboardEvent keyValue." aCharacter.
	closeDelimiter := '([{' at: (')]}' indexOf: openDelimiter).
	level := 1.
	[level > 0 and: [here > 2]]
		whileTrue: [hereChar := string at: (here := here - 1).
			hereChar = closeDelimiter
				ifTrue: [level := level - 1.
					level = 0
						ifTrue: [^ self blinkParenAt: here]]
				ifFalse: [hereChar = openDelimiter
						ifTrue: [level := level + 1]]]! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
cancel
	"Cancel the changes made so far to this text"
	morph cancelEdits! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
cancel: aKeyboardEvent
	"Cancel unsubmitted changes."

	self cancel.
	^ true! !

!TextEditor methodsFor: 'settings' stamp: 'AlainPlantec 11/8/2010 22:09'!
caseSensitiveFinds
	^ self class caseSensitiveFinds! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 5/25/2011 17:38'!
changeAlignment
	"Interactively change the alignment of the text currently being edited."
	| reply  |
	reply := UIManager default 
				chooseFrom:  self alignmentChoices 
				values:  self alignmentChoices.	
	reply ifNil: [^self].
	self applyAttribute: (TextAlignment perform: reply).
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 4/23/2012 10:14'!
changeEmphasis

	|  reply code align menuList startIndex attribute choices |
	choices := self emphasisChoices.
	startIndex := self startIndex.
	align := self text alignmentAt: startIndex ifAbsent: [ 0 ].
	code := self text emphasisAt: startIndex.
	menuList := WriteStream on: Array new.
	menuList nextPut: (code isZero ->  'normal' translated).
	menuList nextPutAll: (choices  collect: [:emph|
							(code anyMask: (TextEmphasis perform: emph) emphasisCode)
								-> emph asString translated ]).
						
	menuList nextPut: ((self text attributesAt: startIndex)
							anySatisfy: [ :attr | attr isKern and: [attr kern < 0 ]]) -> 'narrow' translated.
	reply := UIManager default chooseFrom:  choices values: choices.
	reply notNil 
		ifTrue: [ attribute := TextEmphasis perform: reply.
				self applyAttribute:  attribute].
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
changeEmphasis: aKeyboardEvent
	"Change the emphasis of the current selection or prepare to accept characters with the change in emphasis. Emphasis change amounts to a font change.  Keeps typeahead."

	"control 0..9 -> 0..9"

	| keyCode attribute oldAttributes index thisSel colors |
	keyCode := ('0123456789-=' indexOf: aKeyboardEvent keyCharacter ifAbsent: [1]) - 1.
	oldAttributes := self text attributesAt: self pointIndex.
	thisSel := self selection.

	"Decipher keyCodes for Command 0-9..."
	"
	(keyCode between: 1 and: 5) 
		ifTrue: [attribute := TextFontChange fontNumber: keyCode].
	"
	keyCode = 6 
		ifTrue: [
			colors := #(#black #magenta #red #yellow #green #blue #cyan #white).
			index := UIManager default chooseFrom:  colors title: 'choose color...'.
			index = 0 ifTrue: [^true].
			index <= colors size 
				ifTrue: [attribute := TextColor color: (Color perform: (colors at: index))]
				ifFalse: [
					index := index - colors size - 1.	"Re-number!!!!!!"
					index = 0 ifTrue: [attribute := self chooseColor].
					thisSel ifNil: [^true]	"Could not figure out what to link to"]].
	(keyCode between: 7 and: 11) 
		ifTrue: [
			aKeyboardEvent shiftPressed 
				ifTrue: [
					keyCode = 10 ifTrue: [attribute := TextKern kern: -1].
					keyCode = 11 ifTrue: [attribute := TextKern kern: 1]]
				ifFalse: [
					attribute := TextEmphasis 
								perform: (#(#bold #italic #narrow #underlined #struckOut) at: keyCode - 6).
					oldAttributes 
						do: [:att | (att dominates: attribute) ifTrue: [attribute turnOff]]]].
	keyCode = 0 ifTrue: [ attribute := TextEmphasis normal ].
	attribute ifNotNil: [
		self applyAttribute: attribute].
	^true! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 4/23/2012 08:55'!
changeEmphasisOrAlignment
	| aList reply  code align menuList startIndex attribute |
	startIndex := self startIndex.
	align := self text alignmentAt: startIndex ifAbsent: [ 0 ].
	code := self text emphasisAt: startIndex.
	menuList := WriteStream on: Array new.
	menuList nextPut: (code isZero -> 'normal' translated).
	menuList nextPutAll: (#(bold italic underlined struckOut) collect: [:emph|
		(code anyMask: (TextEmphasis perform: emph) emphasisCode) -> emph asString translated]).
	
	menuList nextPut: ((self text attributesAt: startIndex)
		anySatisfy: [ :attr | attr isKern and: [attr kern < 0 ]]) -> 'narrow' translated.
		
	self alignmentChoices collectWithIndex: [ :type :i |
		menuList nextPut: (align = (i-1)) -> type asString translated ].
	
	aList := #(normal bold italic underlined struckOut narrow leftFlush centered rightFlush justified).
	reply := UIManager default chooseFrom:  aList values: aList.
	reply notNil ifTrue: [
		(self alignmentChoices includes: reply)
			ifTrue: [ attribute :=TextAlignment perform: reply]
			ifFalse: [ attribute := TextEmphasis perform: reply].
		self applyAttribute:  attribute].
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
changeLfToCr: aKeyboardEvent
	"Replace all LFs by CRs.
	Triggered by Cmd-U -- useful when getting code from FTP sites
	jmv- Modified to als change crlf by cr"
	
	| fixed |
	fixed := self selection string.
	fixed := fixed copyReplaceAll: String crlf with: String cr.
	fixed := fixed copyReplaceAll: String lf with: String cr. 
	self replaceSelectionWith: (Text fromString: fixed).
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 4/21/2012 16:48'!
changeStyle
	"Let user change styles for the current text pane."

	| names reply style current menuList |
	current := TextStyle default copy.
	names := TextStyle actualTextStyles keys.
	menuList := names collect: [ :styleName |
		(styleName = current name) -> styleName ].
	reply := UIManager default chooseFrom:  menuList values: names.	
	reply ifNotNil: [
		(style := TextStyle named: reply) ifNil: [Beeper beep. ^ true].
		self text style: style.
		self paragraph textStyle: style.
		self paragraph composeAll.
		self recomputeSelection].
	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
changeStyle: aKeyboardEvent 
	"Put up the style-change menu"

	self closeTypeIn.
	self changeStyle.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 5/25/2011 18:01'!
changeTextFont
	"Present a menu of available fonts, and if one is chosen, apply it to the current selection."
	
	| curFont newFont attr startIndex |
	startIndex := self startIndex.
	curFont := self text fontAt: startIndex.
	newFont := StrikeFont fromUser: curFont.
	newFont ifNil: [ ^self ].
	attr := TextFontReference toFont: newFont.
	self applyAttribute: attr! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
chooseAlignment
	self changeAlignment! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
chooseColor
	"Make a new Text Color Attribute, let the user pick a color, and return the attribute"

	| attribute |	
	attribute := TextColor color: Color black "default".
	UIManager default chooseColor
		ifNotNil: [:nc | attribute color: nc].
	^ attribute
! !

!TextEditor methodsFor: 'parenblinking' stamp: 'AlainPlantec 11/8/2010 22:09'!
clearParens
	| lastLocation |
	(lastLocation := self editingState lastParenLocation) 
		ifNil: [^ self].
	self text size >= lastLocation 
		ifTrue: [ self text
			removeAttribute: TextEmphasis bold
			from: lastLocation
			to: lastLocation].
	self editingState lastParenLocation: nil! !

!TextEditor methodsFor: 'undo-redo private' stamp: 'AlainPlantec 11/24/2010 16:31'!
clearUndoManager: aKeyboardEvent
	^ self editingState clearUndoManager: aKeyboardEvent.
! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 11/8/2010 22:09'!
closeTypeIn
	"See comment in openTypeIn. It is important to call
	closeTypeIn before executing any non-typing key, making a new selection, etc. It is
	called automatically for menu commands."
	self startOfTyping 
		ifNotNil: [self addTypeInUndoRecord.
			self doneTyping]! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 10/20/2011 00:28'!
codeCompletionAround:  aBlock textMorph:  aTextMorph keyStroke: anEvent
	" No code completion for simple text editor, so nothing special to do here around the block execution"
	aBlock value! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
compareToClipboard
	"Check to see if whether the receiver's text is the same as the text currently on the clipboard, and inform the user."
	| s1 s2 |
	s1 := self clipboardText string.
	s2 := self string.
	s1 = s2 ifTrue: [^ self inform: 'Exact match'].

	(StringHolder new textContents:
		(TextDiffBuilder buildDisplayPatchFrom: s1 to: s2))
		openLabel: 'Comparison to Clipboard Text'! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
compareToClipboard: aKeyboardEvent
	"Compare the receiver to the text on the clipboard."

	self compareToClipboard.
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
copyHiddenInfo
	"In TextLinks, TextDoits, TextColor, and TextURLs, there is hidden
info.  Copy that to the clipboard.  You can paste it and see what it is.
Usually enclosed in <>."

	^ self clipboardTextPut: self hiddenInfo asText! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
copySelection
	"Copy the current selection and store it in the paste buffer, unless a caret.  Undoer & Redoer: undoCutCopy"

	self lineSelectAndEmptyCheck: [^ self].
	self clipboardTextPut: self selection.
	self editingState previousInterval: self selectionInterval selection: self selection.
! !

!TextEditor methodsFor: 'new selection' stamp: 'MarcusDenker 2/27/2012 13:18'!
correctFrom: start to: stop with: aString
	"Make a correction in the model that the user has authorised from somewhere else in the system (such as from the compilier).  
	The user's selection is not changed, only corrected."
	| wasShowing userSelection delta loc |
	
	userSelection := self selectionInterval.
	self selectInvisiblyFrom: start to: stop.
	self replaceSelectionWith: aString asText.
	delta := aString size - (stop - start + 1).
	self selectInvisiblyFrom:
		userSelection first + (userSelection first > start ifFalse: [ 0 ] ifTrue: [ delta ])
		to: userSelection last + (userSelection last > start ifFalse: [ 0 ] ifTrue: [ delta ]).
! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
crWithIndent: aKeyboardEvent
	"Replace the current text selection with CR followed by as many tabs
	as on the current line (+/- bracket count) -- initiated by Shift-Return."
	| char s i tabCount  b |
	self closeTypeIn.
	b := self paragraph characterBlockAtPoint: aKeyboardEvent cursorPoint.
	s := self string.
	i := self stopIndex.
	tabCount := 0.
	[(i := i-1) > 0 and: [(char := s at: i) ~= Character cr]]
		whileTrue:  "Count tabs and brackets (but not a leading bracket)"
		[(char = Character tab and: [i < s size and: [(s at: i+1) ~= $[ ]]) ifTrue: [tabCount := tabCount + 1].
		char = $[ ifTrue: [tabCount := tabCount + 1].
		char = $] ifTrue: [tabCount := tabCount - 1]].
	 "Now inject CR with tabCount tabs"
	self addString: (String streamContents: [ :strm | strm crtab: tabCount ]).
	^ false! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
currentCharacterStyleOrNil
	"Answer the CharacterStyle for the current selection or cursor location if any"

	^self hasSelection
		ifTrue: [ self text characterStyleOrNilAt: self startIndex ]
		ifFalse: [ self text characterStyleOrNilIfApplying: self emphasisHere ]! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
currentParagraphStyle
	"Answer the TextStyle for the current selection or cursor location if any"
	
	| i t |
	i := self startIndex.
	t := self text.
	^i > t size
		ifTrue: [ self text styleIfApplying: self emphasisHere ]
		ifFalse: [ self text textStyleAt: i ]! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
cursorEnd: aKeyboardEvent 

	"Private - Move cursor end of current line."
	| string |
	self closeTypeIn.
	string := self string.
	self
		moveCursor:
			[:position | self walkAlongDisplayedLine
				ifTrue:[| targetLine |
					targetLine := self paragraph lines at: (self paragraph lineIndexOfCharacterIndex: position).
					targetLine = self paragraph lines  last 
						ifTrue:[targetLine last + 1]
						ifFalse:[targetLine last]]
				ifFalse:[
					string
						indexOf: Character cr
						startingAt: position
						ifAbsent:[string size + 1]]]
		forward: true
		specialBlock:[:dummy | string size + 1]
		event: aKeyboardEvent.
	^true! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
cursorHome: aKeyboardEvent

	"Private - Move cursor from position in current line to beginning of
	current line. If control key is pressed put cursor at beginning of text"

	| string |
	string := self string.
	self
		moveCursor: [ :position | self walkAlongDisplayedLine
				ifTrue: [ (self paragraph lines at:(self paragraph lineIndexOfCharacterIndex: position)) first]
				ifFalse: [(string
						lastIndexOf: Character cr
						startingAt: position - 1
						ifAbsent:[0]) + 1]]
		forward: false
		specialBlock: [:dummy | 1]
		event: aKeyboardEvent.
	^true! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
cut
	"Cut out the current selection and redisplay the paragraph if necessary.  Undoer & Redoer: undoCutCopy:"

	self lineSelectAndEmptyCheck: [^ self].
	self clipboardTextPut: self selection.
	self replaceSelectionWith: self nullText.
! !

!TextEditor methodsFor: 'typing support' stamp: 'CamilloBruni 2/26/2011 16:07'!
dispatchCommandOn: aKeyboardEvent return: return
	|asciiValue honorCommandKeys char|

	asciiValue := aKeyboardEvent keyValue.
	honorCommandKeys := self cmdKeysInText.
	"Special keys overwrite crtl+key combinations - at least on Windows. To resolve this
	conflict, assume that keys other than cursor keys aren't used together with Crtl." 
	((self class specialShiftCmdKeys includes: asciiValue) and: [ asciiValue < 27])
		ifTrue: [
			^ self performCmdActionsWith: aKeyboardEvent 
				shifted: aKeyboardEvent controlKeyPressed
				return: return].

	"backspace, and escape keys (ascii 8 and 27) are command keys"
	((honorCommandKeys and: [aKeyboardEvent commandKeyPressed]) or: [self class specialShiftCmdKeys includes: asciiValue]) 
		ifTrue: [
			^ self performCmdActionsWith: aKeyboardEvent 
				shifted: aKeyboardEvent shiftPressed
				return: return].

	"the control key can be used to invoke shift-cmd shortcuts"
	(honorCommandKeys and: [aKeyboardEvent controlKeyPressed])
		ifTrue: [
			^ self performCmdActionsWith: aKeyboardEvent shifted: true return: return].
			
	^ false
! !

!TextEditor methodsFor: 'typing support' stamp: 'CamilloBruni 2/26/2011 14:15'!
dispatchEnterOn: aKeyboardEvent return: aBlock
	aKeyboardEvent keyValue = 13
		 ifFalse: [ ^ false ].
	aKeyboardEvent controlKeyPressed 
		ifTrue: [ aBlock value: (self normalCharacter: aKeyboardEvent) ].
	aKeyboardEvent shiftPressed 
		ifTrue: [  aBlock value:  (self lf: aKeyboardEvent) ].
	aKeyboardEvent commandKeyPressed 
		ifTrue: [  aBlock value:  (self crlf: aKeyboardEvent) ].
	 aBlock value:  (self crWithIndent: aKeyboardEvent)! !

!TextEditor methodsFor: 'typing support' stamp: 'CamilloBruni 2/26/2011 16:40'!
dispatchOn: aKeyboardEvent
	"Carry out the action associated with this character, if any.
	Type-ahead is passed so some routines can flush or use it."
	"TODO rename to dispatch: "
	
	| char return keyEvents|
 	return := [:val| ^ val].

	self acceptCR: aKeyboardEvent return: return.
	self clearParens.
	self dispatchEnterOn: aKeyboardEvent return: return.
	self dispatchCommandOn: aKeyboardEvent return: return.
			
	char := aKeyboardEvent keyCharacter.
      ( ')]}'  includes: char) 
		ifTrue: [ self blinkPrevParen: char ].
		
       ^ self normalCharacter: aKeyboardEvent.
! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'CamilloBruni 2/6/2012 11:38'!
doAgainMany: aKeyboardEvent 
	"Do the previous thing again repeatedly."

	self closeTypeIn.
	^ true! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 11/8/2010 22:09'!
doneTyping
	self startOfTyping: nil! !

!TextEditor methodsFor: 'events' stamp: 'JuanVuletich 10/26/2010 14:23'!
doubleClick: evt

	self selectWord! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
duplicate: aKeyboardEvent
	"Paste the current selection over the prior selection, if it is non-overlapping and
	 legal.  Undoer & Redoer: undoAndReselect."

	self closeTypeIn.
	(self hasSelection and: [self isDisjointFrom: self editingState previousInterval])
		ifTrue: "Something to duplicate"
			[self replace: self editingState previousInterval with: self selection and:
				[self selectAt: self pointIndex]]
		ifFalse:
			[morph flash].
	^true! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
editingState
	^ editingState ifNil: [editingState := self editingStateClass new]! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/9/2010 15:05'!
editingStateClass
	^ EditingState! !

!TextEditor methodsFor: 'constants' stamp: 'StephaneDucasse 5/25/2011 18:22'!
emphasisChoices
	"Return the list of emphasis that are possible to apply on the currently edited text morph"
	
	^  #(normal bold italic narrow underlined struckOut).	
	! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
emphasisHere
	^ self editingState emphasisHere! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
enclose: aKeyboardEvent
	"Insert or remove bracket characters around the current selection."

	| left right startIndex stopIndex oldSelection which text |
	self closeTypeIn.
	startIndex := self startIndex.
	stopIndex := self stopIndex.
	oldSelection := self selection.
	which := '([<{"''' indexOf: aKeyboardEvent keyCharacter ifAbsent: [ ^true ].
	left := '([<{"''' at: which.
	right := ')]>}"''' at: which.
	text := self text.
	((startIndex > 1 and: [stopIndex <= text size])
			and: [ (text at: startIndex-1) = left and: [(text at: stopIndex) = right]])
		ifTrue: [
			"already enclosed; strip off brackets"
			self selectFrom: startIndex-1 to: stopIndex.
			self replaceSelectionWith: oldSelection]
		ifFalse: [
			"not enclosed; enclose by matching brackets"
			self replaceSelectionWith:
				(Text string: (String with: left), oldSelection string, (String with: right) attributes: self emphasisHere).
			self selectFrom: startIndex+1 to: stopIndex].
	^true! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 2/21/2011 19:28'!
encompassParagraph: anInterval
	"Return an interval that includes anInterval, and that comprises one or several whole paragraphs in the receiver.
	Answer starts at the position following a cr (or eventually 1) and ends at a cr (or eventually at self size)"
	^ self string encompassParagraph: anInterval
! !

!TextEditor methodsFor: 'do-its' stamp: 'FernandoOlivero 6/9/2011 23:49'!
evaluateSelectionAndDo: aBlock
	"Treat the current selection as an expression; evaluate it and invoke aBlock with the result."
	| result rcvr ctxt |
	self lineSelectAndEmptyCheck: [^ nil].

	(model respondsTo: #doItReceiver) 
		ifTrue: [ rcvr := model doItReceiver.
				ctxt := model doItContext]
		ifFalse: [rcvr := ctxt := nil].
	result := [
		rcvr class evaluatorClass new 
			evaluate: self selectionAsStream
			in: ctxt
			to: rcvr
			notifying: self
			ifFail: [morph flash. ^ nil]
			logged: true.
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
	^aBlock value: result! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
exchange
	"See comment in exchangeWith:"

	self exchangeWith: self editingState previousInterval! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
exchange: aKeyboardEvent
	"Exchange the current and prior selections."

	self closeTypeIn.
	self exchange.
	^true! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 12/14/2010 14:45'!
exchangeWith: prior
	"If the prior selection is non-overlapping and legal, exchange the text of
	 it with the current selection and leave the currently selected text selected
	 in the location of the prior selection (or leave a caret after a non-caret if it was
	 exchanged with a caret).  If both selections are carets, flash & do nothing.
	 Don't affect the paste buffer.  Undoer: itself; Redoer: Undoer."

	| start stop before currSelection priorSelection delta redoArgs altInterval undoArgs undoRec |
	start := self startIndex.
	stop := self stopIndex - 1.
	((prior first <= prior last) | (start <= stop) "Something to exchange" and:
			[self isDisjointFrom: prior])
		ifTrue:
			[before := prior last < start.
			currSelection := self selection.
			priorSelection := self text copyFrom: prior first to: prior last.
			delta := before ifTrue: [0] ifFalse: [priorSelection size - currSelection size].
			self zapSelectionWith: priorSelection.
			redoArgs := { prior. start to: stop}.
						
			self selectFrom: prior first + delta to: prior last + delta.
			delta := before ifTrue: [stop - prior last] ifFalse: [start - prior first].
			self zapSelectionWith: currSelection.
			
			altInterval := prior first + delta to: prior last + delta.
			undoArgs := {altInterval. self startIndex to: self stopIndex - 1}.
			"self undoer: #exchangeWith: with: altInterval."
			prior first > prior last ifTrue: [self selectAt: self editingState previousInterval last + 1].
			self editingState
				redoArray: { morph. #undoRedoExchange:with:. redoArgs} 
				undoArray: {morph. #undoRedoExchange:with:. undoArgs}]
		ifFalse:
			[morph flash]! !

!TextEditor methodsFor: 'undo-redo private' stamp: 'AlainPlantec 11/8/2010 22:09'!
exploreUndoManager: aKeyboardEvent
	^ self editingState exploreUndoManager: aKeyboardEvent! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/11/2010 21:47'!
find
	"Prompt the user for a string to search for, and search the receiver from the current selection onward for it."
	morph openFindDialog! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
find: aKeyboardEvent
	"Prompt the user for what to find, then find it, searching from the current selection onward.  1/24/96 sw"

	self closeTypeIn.
	self find.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/17/2010 16:00'!
findAgain
	"Find the text-to-find again."
	| where |
	where := self findReplaceService findNextSubString: self findReplaceService findString inTextMorph: self morph.
	where isNil
		ifTrue: [self flash]
		ifFalse: [ self selectInterval: where].
	^ where
! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
findAgain: aKeyboardEvent 
	"Find the desired text again.  1/24/96 sw"
	self closeTypeIn.
	self findAgain.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/16/2010 13:19'!
findAgainAndReplace
	"Find the text-to-find again and eventually replace it with the current replace text."
	| where |
	where := self findAgain.
	where isNil
		ifFalse: [self findReplaceService replaceText 
				ifNotEmpty: [:rt | self replaceSelectionWith: rt]]
! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/9/2010 19:40'!
findAll: aRegex endingAt: searchIdx 
	| ranges |
	ranges := aRegex matchingRangesIn: (self string copyFrom: 1 to: searchIdx).
	^ ranges 
			! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/9/2010 17:32'!
findAll: aRegex startingAt: searchIdx 
	| ranges |
	ranges := aRegex matchingRangesIn: (self string copyFrom: searchIdx to: self string size).
	^ ranges collect: [:i | i + searchIdx - 1]
			! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/9/2010 18:40'!
findAndSelect: aRegex startingAt: anIndex searchBackwards: searchBackwards
	| oldSelectionInterval where |
	self closeTypeIn.
	oldSelectionInterval := self selectionInterval.
	self selectInvisiblyFrom: anIndex to: anIndex - 1.
	where :=  searchBackwards
		ifTrue: [self findAndSelectPreviousOccurenceOf: aRegex]
		ifFalse: [self findAndSelectNextOccurenceOf: aRegex].
	where ifNil: [self selectInterval: oldSelectionInterval].
	^ where! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/19/2010 17:49'!
findAndSelectNextOccurenceOf: aRegex
	| where |
	where := self findNext: aRegex startingAt: self stopIndex.
	where ifNotNil: [self selectInterval: where].
	"(morph respondsTo: #editView) ifTrue: [morph editView selectionInterval: self selectionInterval]."
	^ where! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/9/2010 19:35'!
findAndSelectPreviousOccurenceOf: aRegex
	| where |
	where := self findPrevious: aRegex startingAt: self startIndex.
	where isNil ifFalse: [self selectInterval: where].
	(morph respondsTo: #editView) ifTrue: [morph editView selectionInterval: self selectionInterval].
	^ where! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/19/2010 17:26'!
findNext: aRegex startingAt: searchIdx 
	| strm match range |
	strm := (self string copyFrom: searchIdx to: self string size) readStream.
	aRegex searchStream: strm.
	match := aRegex subexpression: 1.
	match ifNotNil: [range := (aRegex position + searchIdx - match size) to: (aRegex position + searchIdx - 1)].
	^ range! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/20/2010 08:45'!
findNextString: aSubString startingAt: searchIdx 
	| idx |
	idx := self string findString: aSubString startingAt: searchIdx.
	^ idx isZero ifFalse: [idx to: idx + aSubString size - 1]
			! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/9/2010 19:40'!
findPrevious: aRegex startingAt: searchIdx 
	| allRanges |
	allRanges := self findAll: aRegex endingAt: searchIdx.
	^ allRanges notEmpty
		ifTrue: [allRanges last].
! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/12/2010 08:27'!
findRegex
	^ self findReplaceService findRegex! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/12/2010 21:07'!
findReplaceService
	^ FindReplaceService default! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/12/2010 08:27'!
findText
	^ self findReplaceService findText! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/17/2010 16:05'!
findText: aStringOrText
	self findReplaceService findText: aStringOrText ! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/26/2010 18:13'!
findText: aString isRegex: aBoolean
	self findReplaceService findText: aString isRegex: aBoolean..
	self findReplaceService changed: #newFinding
! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/12/2010 12:09'!
findText: aString isRegex: isRegex caseSensitive: caseSensitive entireWordsOnly: entireWordsOnly
	self findReplaceService findText: aString isRegex: isRegex caseSensitive: caseSensitive entireWordsOnly: entireWordsOnly
! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/12/2010 10:47'!
findText: aString isRegex: isRegex entireWordsOnly: entireWordsOnly
	self findReplaceService findText: aString isRegex: isRegex entireWordsOnly: entireWordsOnly
! !

!TextEditor methodsFor: 'displaying' stamp: 'AlainPlantec 11/8/2010 22:09'!
flash
	^ morph flash! !

!TextEditor methodsFor: 'events' stamp: 'AlainPlantec 11/19/2010 10:46'!
focused: aBoolean
	self editingState focused: aBoolean from: morph.
! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 12/14/2010 22:49'!
forwardDelete: aKeyboardEvent
	"Delete forward over the next character"
	self closeTypeIn.
	self hasSelection 
		ifFalse: [	| idx1 idx2 |
				idx1 := self startIndex min: self stopIndex.
				idx2 := self stopIndex max: self startIndex.
				aKeyboardEvent shiftPressed
					ifTrue: [idx2 := (self nextWord: idx1)]
					ifFalse: [idx2 := idx2 + 1].
				self selectInvisiblyFrom: idx1 to: idx2 - 1].
	self addDeleteSelectionUndoRecord.
	self zapSelectionWith: self nullText.
	^ false
! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
getPluggableYellowButtonMenu: shiftKeyState
	| customMenu |
	^ (morph notNil and: [(customMenu := morph getMenu: shiftKeyState) notNil])
		ifTrue: [customMenu]
		ifFalse:
			[shiftKeyState
				ifTrue: [self class shiftedYellowButtonMenu]
				ifFalse: [self class yellowButtonMenu]]! !

!TextEditor methodsFor: 'typing support' stamp: 'GuillermoPolito 4/20/2011 19:36'!
handleEditionAction: anAction fromKeyboardEvent: aKeyboardEvent
	(self editingState handlesKeyboard: aKeyboardEvent from: morph)
		ifFalse: [self handleKeystrokeAction: anAction ].
	self editingState keystroke: aKeyboardEvent from: morph.! !

!TextEditor methodsFor: 'typing support' stamp: 'GuillermoPolito 4/20/2011 18:44'!
handleKeystrokeAction: aBlock
	(aBlock value)
			ifTrue: [self doneTyping.
			self storeSelectionInParagraph.
			^self].
	self openTypeIn.
	self hasSelection 
		ifTrue: [ "save highlighted characters"
			self editingState previousInterval: self selectionInterval selection: self selection].
	self editingState unselect.
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
hasCaret
	^self editingState markBlock = self editingState pointBlock! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
hiddenInfo
	"In TextLinks, TextDoits, TextColor, and TextURLs, there is hidden info.  Return the entire string that was used by Cmd-6 to create this text attribute.  Usually enclosed in < >."

	| attrList |
	attrList := self text attributesAt: (self pointIndex + self markIndex)//2.
	attrList do: [:attr |
		attr forTextActionInfoDo: [ :info |
			^ self selection asString, '<', info, '>']].
	"If none of the above"
	attrList do: [:attr |
		attr forTextColorDo: [ :color |
			^ self selection asString, '<', color printString, '>']].
	^ self selection asString, '[No hidden info]'! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
inOutdent: aKeyboardEvent delta: delta
	"Add/remove a tab at the front of every line occupied by the selection. 
	Derived from work by Larry Tesler back in December 1985.  Now triggered by Cmd-L and Cmd-R.  2/29/96 sw"

	| cr realStart realStop lines startLine stopLine start stop adjustStart indentation size numLines inStream newString outStream |
	cr := Character cr.

	"Operate on entire lines, but remember the real selection for re-highlighting later"
	realStart := self startIndex.
	realStop := self stopIndex - 1.

	"Special case a caret on a line of its own, including weird case at end of paragraph"
	(realStart > realStop and: [
				realStart < 2 or: [(self string at: realStart - 1) == cr]])
		ifTrue:
			[delta < 0
				ifTrue: [
					morph flash]
				ifFalse: [
					self replaceSelectionWith: Character tab asSymbol asText.
					self selectAt: realStart + 1].
			^true].

	lines := self paragraph lines.
	startLine := self paragraph lineIndexOfCharacterIndex: realStart.
	stopLine := self paragraph lineIndexOfCharacterIndex: (realStart max: realStop).
	start := (lines at: startLine) first.
	stop := (lines at: stopLine) last.
	
	"Pin the start of highlighting unless the selection starts a line"
	adjustStart := realStart > start.

	"Find the indentation of the least-indented non-blank line; never outdent more"
	indentation := (startLine to: stopLine) inject: 1000 into: [ :previousValue :each |
		previousValue min: (self paragraph indentationOfLineIndex: each ifBlank: [ :tabs | 1000 ])].			

	size :=  stop + 1 - start.
	numLines := stopLine + 1 - startLine.
	inStream := ReadStream on: self string from: start to: stop.

	newString := String new: size + ((numLines * delta) max: 0).
	outStream := ReadWriteStream on: newString.

	"This subroutine does the actual work"
	self indent: delta fromStream: inStream toStream: outStream.

	"Adjust the range that will be highlighted later"
	adjustStart ifTrue: [realStart := (realStart + delta) max: start].
	realStop := realStop + outStream position - size.

	"Prepare for another iteration"
	indentation := indentation + delta.
	size := outStream position.
	inStream := outStream setFrom: 1 to: size.

	outStream == nil
		ifTrue: 	"tried to outdent but some line(s) were already left flush"
			[morph flash]
		ifFalse:
			[self selectInvisiblyFrom: start to: stop.
			size = newString size ifFalse: [newString := outStream contents].
			self replaceSelectionWith: newString asText].
	self selectFrom: realStart to: realStop. 	"highlight only the original range"
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
indent: aKeyboardEvent
	"Add a tab at the front of every line occupied by the selection. Invoked from keyboard via cmd-shift-R.  2/29/96 sw"

	^ self inOutdent: aKeyboardEvent delta: 1! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
indent: delta fromStream: inStream toStream: outStream
	"Append the contents of inStream to outStream, adding or deleting delta or -delta
	 tabs at the beginning, and after every CR except a final CR.  Do not add tabs
	 to totally empty lines, and be sure nothing but tabs are removed from lines."

	| ch skip cr tab prev atEnd |
	cr := Character cr.
	tab := Character tab.
	delta > 0
		ifTrue: "shift right"
			[prev := cr.
			 [ch := (atEnd := inStream atEnd) ifTrue: [cr] ifFalse: [inStream next].
			  (prev == cr and: [ch ~~ cr]) ifTrue:
				[delta timesRepeat: [outStream nextPut: tab]].
			  atEnd]
				whileFalse:
					[outStream nextPut: ch.
					prev := ch]]
		ifFalse: "shift left"
			[skip := delta. "a negative number"
			 [inStream atEnd] whileFalse:
				[((ch := inStream next) == tab and: [skip < 0]) ifFalse:
					[outStream nextPut: ch].
				skip := ch == cr ifTrue: [delta] ifFalse: [skip + 1]]]! !

!TextEditor methodsFor: 'initialize-release' stamp: 'IgorStasenko 12/12/2010 04:06'!
initWithMorph: aMorph
	
	super initialize.
	morph := aMorph.
	self resetState.! !

!TextEditor methodsFor: 'new selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
insertAndSelect: aString at: anInteger

	self replace: (anInteger to: anInteger - 1)
		with: (Text string: (' ' , aString)
					attributes: self emphasisHere)
		and: [self ]! !

!TextEditor methodsFor: 'shortcuts' stamp: 'GuillermoPolito 5/21/2012 01:30'!
installKeymappingsOn: aMorph
	aMorph attachKeymapCategory: #TextEditor! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
isDisjointFrom: anInterval
	"Answer true if anInterval is a caret not touching or within the current
	 interval, or if anInterval is a non-caret that does not overlap the current
	 selection."

	| fudge |
	fudge := anInterval size = 0 ifTrue: [1] ifFalse: [0].
	^(anInterval last + fudge < self startIndex or:
			[anInterval first - fudge >= self stopIndex])
! !

!TextEditor methodsFor: 'testing' stamp: 'AlainPlantec 11/8/2010 22:09'!
isTextEditor
	^ true ! !

!TextEditor methodsFor: 'typing support' stamp: 'GuillermoPolito 4/20/2011 19:42'!
keystroke: aKeyboardEvent
	self handleEditionAction: [ self dispatchOn: aKeyboardEvent ] 
		fromKeyboardEvent: aKeyboardEvent.! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/17/2010 18:47'!
lastFont	
	"Answer the Font for to be used if positioned at the end of the text"
	| t |	
	t := self text.	
	^ t fontAt: t size + 1! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
lastParagraphStyle
	"Answer the TextStyle for to be used if positioned at the end of the text"
	
	| t |
	t := self text.
	^self startIndex > t size
		ifTrue: [ self text styleIfApplying: self emphasisHere ]
		ifFalse: [ self text textStyleAt: t size +1 ]! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
makeCapitalized: aKeyboardEvent
	"Force the current selection to uppercase.  Triggered by Cmd-z."

	| prev |
	self closeTypeIn.
	prev := $-.  "not a letter"
	self replaceSelectionWith: (Text fromString:
		(self selection string collect:
			[:c | prev := prev isLetter ifTrue: [c asLowercase] ifFalse: [c asUppercase]])).
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
makeLowercase: aKeyboardEvent
	"Force the current selection to lowercase.  Triggered by Cmd-X."

	self closeTypeIn.
	self replaceSelectionWith: (Text fromString: (self selection string asLowercase)).
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
makeUppercase: aKeyboardEvent
	"Force the current selection to uppercase.  Triggered by Cmd-Y."

	self closeTypeIn.
	self replaceSelectionWith: (Text fromString: (self selection string asUppercase)).
	^ true! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/16/2010 10:07'!
markBlock
	^ self editingState markBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
markIndex
	^ self editingState markBlock stringIndex! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
markIndex: anIndex
	self editingState markBlock: (self paragraph characterBlockForIndex: anIndex)
! !

!TextEditor methodsFor: 'accessing' stamp: 'MarcusDenker 12/9/2010 14:51'!
model
	^model! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
model: aModel 
	"Controller|model: and Controller|view: are sent by View|controller: in 
	order to coordinate the links between the model, view, and controller. In 
	ordinary usage, the receiver is created and passed as the parameter to 
	View|controller: so that the receiver's model and view links can be set 
	up by the view."

	model := aModel! !

!TextEditor methodsFor: 'events' stamp: 'CamilloBruni 2/6/2012 23:16'!
mouseDown: event 
	"An attempt to break up the old processRedButton code into three phases"
	
	self editingState mouseDown: event from: morph.
		
	self closeTypeIn.
	
	"no matter what, if shift is pressed extend the selection"
	event shiftPressed
		ifTrue: [ ^ self mouseMove: event ].
	
	event yellowButtonPressed 
		ifTrue: [ self yellowButtonDown: event ]
		ifFalse: [ 
			"no right/yellowbutton click just reposoition the cursor"
			self updateSelectionFromEvent: event ].! !

!TextEditor methodsFor: 'events' stamp: 'CamilloBruni 2/1/2012 20:11'!
mouseMove: evt 
	"Change the selection in response to mouse-down drag"

	self editingState pointBlock: (self paragraph characterBlockAtPoint: (evt cursorPoint)).
	
	self paragraph move: evt for: model controller: self.
			
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'events' stamp: 'CamilloBruni 2/6/2012 23:02'!
mouseUp: evt
	"An attempt to break up the old processRedButton 
	code into three phases"
	| mouseDownInterval clickPoint b|
	
	(mouseDownInterval := self editingState mouseDownInterval)
		ifNil: [^ self]. 
		
	(self paragraph click: evt for: model controller: self ) 
		ifTrue: [
			self updateSelectionFromEvent: evt.
			evt hand releaseKeyboardFocus: self.
			^ self ].

	(evt yellowButtonChanged or: [ evt commandKeyPressed]) 
		ifTrue: [
			(self morph editView yellowButtonActivity: evt shiftPressed) 
				ifTrue: [ ^ self ]].
		
	(self hasCaret and: [mouseDownInterval = self selectionInterval])
		ifTrue: [self selectWord].
		
	self setEmphasisHereFromText.
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
moveCursor: directionBlock forward: forward specialBlock: specialBlock event: aKeyboardEvent

	super moveCursor: directionBlock forward: forward specialBlock: specialBlock event: aKeyboardEvent.
	self setEmphasisHereFromTextForward: forward! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/12/2010 21:38'!
normalCharacter: aKeyboardEvent 
	"A nonspecial character is to be added to the stream of characters."

	self hasSelection 
		ifTrue: [self addDeleteSelectionUndoRecord].
	^ super normalCharacter: aKeyboardEvent.
! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
nullText

	^Text string: '' attributes: self emphasisHere! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
offerFontMenu
	"Present a menu of available fonts, and if one is chosen, apply it to the current selection.  
	Use only names of Fonts of this paragraph  "
	
	^self changeTextFont! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
offerFontMenu: aKeyboardEvent 
	"The user typed the command key that requests a font change; Offer the font menu.  5/27/96 sw"

	self closeTypeIn.
	self offerFontMenu.
	^ true! !

!TextEditor methodsFor: 'deprecated' stamp: 'AlainPlantec 11/8/2010 22:09'!
offerMenuFromEsc: aKeyboardEvent
	"The escape key was hit while the receiver has the keyboard focus; take action"
 	ActiveEvent shiftPressed 
		ifFalse: [ self raiseContextMenu: aKeyboardEvent].
	^ false 	! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 11/8/2010 22:09'!
openTypeIn
	"Set up startOfTyping to keep track of the leftmost backspace.
	 You can't undo typing until after closeTypeIn."

	self startOfTyping 
		ifNil: [self editingState previousInterval: (1 to: 0) selection: self nullText.
			self startOfTyping: self startIndex]! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
outdent: aKeyboardEvent
	"Remove a tab from the front of every line occupied by the selection.
	Invoked from keyboard via cmd-shift-L.  2/29/96 sw"

	^ self inOutdent: aKeyboardEvent delta: -1! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
pageHeight
	| howManyLines visibleHeight totalHeight ratio |
	howManyLines := self paragraph numberOfLines.
	visibleHeight := self visibleHeight.
	totalHeight := self totalTextHeight.
	ratio := visibleHeight / totalHeight.
	^(ratio * howManyLines) rounded - 2! !

!TextEditor methodsFor: 'private' stamp: 'IgorStasenko 12/12/2010 03:30'!
paragraph
	^ morph 
		ifNotNil: [morph paragraph]
		ifNil: [self error: 'paragraph is not available']! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
pasteRecent
	"Paste an item chose from RecentClippings."

	| clipping |
	(clipping := Clipboard chooseRecentClipping) ifNil: [^ self].
	Clipboard clipboardText: clipping.
	^ self paste! !

!TextEditor methodsFor: 'typing support' stamp: 'CamilloBruni 2/26/2011 14:02'!
performCmdActionsWith: aKeyboardEvent shifted: aBoolean 
	| asciiValue actions|
	asciiValue := aKeyboardEvent keyValue + 1.
	actions := aBoolean
				ifTrue: [self class shiftCmdActions]
				ifFalse: [self class cmdActions].
	^ self perform: (actions at: asciiValue) with: aKeyboardEvent! !

!TextEditor methodsFor: 'typing support' stamp: 'CamilloBruni 2/26/2011 16:40'!
performCmdActionsWith: aKeyboardEvent shifted: aBoolean return: return
	| asciiValue actions action|
	asciiValue := aKeyboardEvent keyValue + 1.
	actions := aBoolean
				ifTrue: [self class shiftCmdActions]
				ifFalse: [self class cmdActions].
	action := (actions at: asciiValue).
	action == #noop: ifTrue: [ ^ false ].
	return value: ( self perform: action with: aKeyboardEvent).! !

!TextEditor methodsFor: 'pluggable menus' stamp: 'CamilloBruni 2/6/2012 11:25'!
pluggableYellowButtonActivity: shiftKeyState
	"Invoke the model's popup menu."

	| menu |
	(menu := self getPluggableYellowButtonMenu: shiftKeyState)
		ifNil: [
			^ false]
		ifNotNil: [
			menu invokeOn: model orSendTo: self.
			^ true]! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/16/2010 10:07'!
pointBlock
	^ self editingState pointBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
pointIndex
	^ self editingState pointBlock stringIndex! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
pointIndex: anIndex
	self editingState pointBlock: (self paragraph characterBlockForIndex: anIndex)
! !

!TextEditor methodsFor: 'typing support' stamp: 'GuillermoPolito 4/20/2011 19:35'!
processKeyStroke: aKeyboardEvent
	"Key struck on the keyboard. Find out which one and, if special, carry 
	out the associated special action. Otherwise, add the character to the 
	stream of characters.  Undoer & Redoer: see closeTypeIn."
	
	^self handleKeystrokeAction: [self dispatchOn: aKeyboardEvent].! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'CamilloBruni 2/4/2012 14:50'!
raiseContextMenu: aKeyboardEvent

	(morph respondsTo: #editView)
		ifTrue: [ ^ morph editView yellowButtonActivity: ActiveEvent shiftPressed].
	^ true! !

!TextEditor methodsFor: 'current selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
recomputeInterval
	"The same characters are selected but their coordinates may have changed."

	self markIndex: self markIndex; pointIndex: self pointIndex! !

!TextEditor methodsFor: 'current selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
recomputeSelection
	"Redetermine the selection according to the start and stop block indices; 
	do not highlight."

	self recomputeInterval! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
redo
	"redo previous edit"

	self editingState redo ifFalse: [morph flash]. 
! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
redo: aKeyboardEvent 
	"Redo the last edit."

	self redo.
	^true! !

!TextEditor methodsFor: 'undoers-redoers' stamp: 'AlainPlantec 11/8/2010 22:09'!
redoTypeIn: aText interval: anInterval
	self selectInterval: anInterval.
	self replace: self selectionInterval with: aText and:
		[self selectInterval: (anInterval first to: anInterval first + aText size - 1)].
! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 12/14/2010 14:46'!
replace: xoldInterval with: newText and: selectingBlock 
	"Replace the text in oldInterval with newText and
	execute selectingBlock to establish the new selection.
	Create an UndoRecord to allow perfect undoing."
	| prevSel currInterval undoRec |
	self selectInterval: xoldInterval.
	prevSel := self selection.
	currInterval := self selectionInterval.
	self editingState previousInterval: currInterval selection: self selection.
	self zapSelectionWith: newText.
	selectingBlock value.
	((prevSel isEmpty and: [newText isEmpty]) and: [currInterval size < 1]) 
		ifFalse: [self editingState
			redoArray: {morph. #redoTypeIn:interval:. {newText. currInterval}} 
			undoArray: {morph. #undoTypeIn:interval:. {prevSel. currInterval first to: currInterval first + newText size - 1}}]! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/11/2010 23:40'!
replaceAll: aRegex with: aText
	| selec ranges |
	self editingState undoManager openRecordGroup.
	selec := self selectionInterval.
	ranges := self findAll: aRegex startingAt: selec first. 
	ranges reverseDo: [:r | self selectInvisiblyFrom: r first to: r last.
		self replaceSelectionWith: aText].
	self selectInterval: selec.
	self editingState undoManager closeRecordGroup.
! !

!TextEditor methodsFor: 'find-select' stamp: 'AlainPlantec 11/11/2010 23:42'!
replaceAll: aRegex with: aText startingAt: startIdx
	| selec ranges |
	self editingState undoManager openRecordGroup.
	selec := self selectionInterval.
	ranges := self findAll: aRegex startingAt: startIdx. 
	ranges reverseDo: [:r | self selectInvisiblyFrom: r first to: r last.
		self replaceSelectionWith: aText].
	self selectInterval: selec.
	self editingState undoManager closeRecordGroup.
! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
replaceSelectionWith: aText
	self replace: self selectionInterval with: aText and: []
! !

!TextEditor methodsFor: 'initialize-release' stamp: 'AlainPlantec 11/11/2010 19:01'!
resetState 
	"Establish the initial conditions for editing the paragraph: place caret 
	before first character, set the emphasis to that of the first character,
	and save the paragraph for purposes of canceling."
 
	self editingState markBlock: self paragraph defaultCharacterBlock.
	self editingState pointBlock: self editingState markBlock copy.
	self startOfTyping: nil.
	self editingState previousInterval: (1 to: 0) selection: self nullText.
	self setEmphasisHereFromText.
	selectionShowing := true! !

!TextEditor methodsFor: 'current selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
reverseSelection
	"Reverse the valence of the current selection highlighting."
	selectionShowing := selectionShowing not.
	self paragraph reverseFrom: self editingState pointBlock to: self editingState markBlock! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
sameColumn: start newLine: lineBlock forward: isForward
	"Private - Compute the index in my text
	with the line number derived from lineBlock,"
	" a one argument block accepting the old line number.
	The position inside the line will be preserved as good as possible"
	"The boolean isForward is used in the border case to determine if
	we should move to the beginning or the end of the line."
	| column currentLine offsetAtTargetLine targetEOL lines numberOfLines currentLineNumber targetLineNumber |
	self walkAlongDisplayedLine
		ifTrue: [lines := self paragraph lines.
			numberOfLines := self paragraph numberOfLines.
			currentLineNumber := self paragraph lineIndexOfCharacterIndex: start.
			currentLine := lines at: currentLineNumber]
		ifFalse: [lines := self lines.
			numberOfLines := lines size.
			currentLine := lines
				detect:[:lineInterval | lineInterval last >= start]
				ifNone:[lines last].
			currentLineNumber := currentLine second].
	column := start - currentLine first.
	targetLineNumber := ((lineBlock value: currentLineNumber) max: 1) min: numberOfLines.
	offsetAtTargetLine := (lines at: targetLineNumber) first.
	targetEOL := (lines at: targetLineNumber) last + (targetLineNumber = numberOfLines ifTrue:[1]ifFalse:[0]).
	targetLineNumber = currentLineNumber
	"No movement or movement failed. Move to beginning or end of line."
		ifTrue:[^isForward
			ifTrue:[targetEOL]
			ifFalse:[offsetAtTargetLine]].
	^offsetAtTargetLine + column min: targetEOL.! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
saveContentsInFile
	"Save the receiver's contents string to a file, prompting the user for a file-name.  Suggest a reasonable file-name."

	| fileName stringToSave parentWindow labelToUse suggestedName lastIndex |
	stringToSave := self string. 
	stringToSave size = 0 ifTrue: [^self inform: 'nothing to save.'].
	parentWindow := model dependents 
				detect: [:dep | dep isKindOf: SystemWindow]
				ifNone: [nil].
	labelToUse := parentWindow ifNil: ['Untitled']
				ifNotNil: [parentWindow label].
	suggestedName := nil.
	#(#('Decompressed contents of: ' '.gz')) do: 
			[:leaderTrailer | 
			"can add more here..."

			(labelToUse beginsWith: leaderTrailer first) 
				ifTrue: 
					[suggestedName := labelToUse copyFrom: leaderTrailer first size + 1
								to: labelToUse size.
					(labelToUse endsWith: leaderTrailer last) 
						ifTrue: 
							[suggestedName := suggestedName copyFrom: 1
										to: suggestedName size - leaderTrailer last size]
						ifFalse: 
							[lastIndex := suggestedName lastIndexOf: $. ifAbsent: [0].
							(lastIndex = 0 or: [lastIndex = 1]) 
								ifFalse: [suggestedName := suggestedName copyFrom: 1 to: lastIndex - 1]]]].
	suggestedName ifNil: [suggestedName := labelToUse , '.text'].
	fileName := UIManager default request: 'File name?'
				initialAnswer: suggestedName.
	fileName isEmptyOrNil 
		ifFalse: 
			[(FileStream newFileNamed: fileName)
				nextPutAll: stringToSave;
				close]! !

!TextEditor methodsFor: 'scrolling' stamp: 'AlainPlantec 11/11/2010 11:05'!
scrollBy: ignore 
	"Ignore scroll requests."
! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'AlainPlantec 11/13/2010 08:23'!
search: aKeyboardEvent
	"Invoked by Ctrl-S.  Same as 'again', but always uses the existing findText
	 from current FindReplaceService"

	self closeTypeIn.
	self findAgain.
	^true! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
selectCurrentTypeIn: aKeyboardEvent 
	"Select what would be replaced by an undo (e.g., the last typeIn)."

	self closeTypeIn.
	self selectInterval: self editingState previousInterval.
	^ true! !

!TextEditor methodsFor: 'new selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
selectFrom: start to: stop
	"Select the specified characters inclusive."
	self selectInvisiblyFrom: start to: stop.
	self closeTypeIn.
	self storeSelectionInParagraph.
	"Preserve current emphasis if selection is empty"
	stop > start 
		ifTrue: [self setEmphasisHereFromTextForward: true ]! !

!TextEditor methodsFor: 'new selection' stamp: 'FernandoOlivero 6/16/2010 12:53'!
selectLine
	"Make the receiver's selection, if it currently consists of an insertion point only, encompass the current line."
	self hasSelection ifTrue:[^self].
	self selectInterval: (self encompassParagraph: self selectionInterval)! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
selection
	"Answer the text in the paragraph that is currently selected."

	^self text copyFrom: self startIndex to: self stopIndex - 1 ! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
selectionAsStream
	"Answer a ReadStream on the text in the paragraph that is currently 
	selected."

	^ReadStream
		on: (self string copyFrom: self startIndex to: self stopIndex - 1)! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/8/2010 22:09'!
setEmphasisHere
	self editingState emphasisHere: ((self text attributesAt: (self pointIndex - 1 max: 1) forStyle: self textStyle)
					select: [:att | att mayBeExtended])! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 11/8/2010 22:09'!
setEmphasisHereFromText

	self setEmphasisHereFromTextForward: true! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 11/8/2010 22:09'!
setEmphasisHereFromTextForward: f

	| i t forward delta prevIsSeparator nextIsSeparator |
	i := self pointIndex.
	t := self text.
	"Try to set emphasisHere correctly after whitespace.
	Most important after a cr, i.e. at the start of a new line"
	prevIsSeparator :=  i > 1 and: [ (t at: i-1) isSeparator ].
	nextIsSeparator := i <= t size and: [ (t at: i) isSeparator ].
	forward := prevIsSeparator = nextIsSeparator
		ifTrue: [ f ]
		ifFalse: [ nextIsSeparator ].
	delta := forward ifTrue: [ 1 ] ifFalse: [ 0 ].
	self editingState emphasisHere: ((t attributesAt: (i - delta max: 1)) select: [:att | att mayBeExtended]).
	! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/27/2010 16:27'!
setSearch: aString
	"Make the current selection, if any, be the current search string."
	self findText: aString isRegex: false.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/10/2010 19:55'!
setSearchString
	"Make the current selection, if any, be the current search string."
	self closeTypeIn.
	self setSearch: self selection string.
	^ true! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
setSearchString: aKeyboardEvent
	"Establish the current selection as the current search string."
	^ self setSearchString! !

!TextEditor methodsFor: 'menu messages' stamp: 'SvenVanCaekenberghe 1/8/2012 14:49'!
setSelectorSearch: aStringOrText 
	"Make the current selection as a selector search in current FindReplaceService."
	| regex |
	regex := aStringOrText asString trimBoth.
	regex := regex copyReplaceAll: '#' with: ''.
	regex := regex copyReplaceAll: ' ' with: ''.
	self findText: regex isRegex: false caseSensitive: true entireWordsOnly: false.
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
shiftEnclose: aKeyboardEvent
	"Insert or remove bracket characters around the current selection."

	| char left right startIndex stopIndex oldSelection which text |
	char := aKeyboardEvent keyCharacter.
	char = $9 ifTrue: [ char := $( ].
	char = $, ifTrue: [ char := $< ].
	char = $[ ifTrue: [ char := ${ ].
	char = $' ifTrue: [ char := $" ].
	char asciiValue = 27 ifTrue: [ char := ${ ].	"ctrl-["

	self closeTypeIn.
	startIndex := self startIndex.
	stopIndex := self stopIndex.
	oldSelection := self selection.
	which := '([<{"''' indexOf: char ifAbsent: [1].
	left := '([<{"''' at: which.
	right := ')]>}"''' at: which.
	text := self text.
	((startIndex > 1 and: [stopIndex <= text size])
			and: [ (text at: startIndex-1) = left and: [(text at: stopIndex) = right]])
		ifTrue: [
			"already enclosed; strip off brackets"
			self selectFrom: startIndex-1 to: stopIndex.
			self replaceSelectionWith: oldSelection]
		ifFalse: [
			"not enclosed; enclose by matching brackets"
			self replaceSelectionWith:
				(Text string: (String with: left), oldSelection string, (String with: right) attributes: self emphasisHere).
			self selectFrom: startIndex+1 to: stopIndex].
	^true! !

!TextEditor methodsFor: 'pluggable menus' stamp: 'jmv 11/4/2008 11:51'!
shiftedTextPaneMenuRequest
	"The user chose the more... branch from the text-pane menu."

	^ self pluggableYellowButtonActivity: true! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/24/2010 16:12'!
startBlock
	^ self editingState pointBlock min: self editingState markBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
startIndex
	^ self startBlock stringIndex! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 11/8/2010 22:09'!
startOfTyping
	^ self editingState startOfTyping  ! !

!TextEditor methodsFor: 'typing support' stamp: 'AlainPlantec 11/8/2010 22:09'!
startOfTyping: anIntegerIndex
	self editingState startOfTyping:  anIntegerIndex! !

!TextEditor methodsFor: 'initialize-release' stamp: 'AlainPlantec 11/8/2010 22:09'!
stateArray
	^ Array with: self editingState! !

!TextEditor methodsFor: 'initialize-release' stamp: 'AlainPlantec 11/8/2010 22:09'!
stateArrayPut: stateArray
	editingState := stateArray at: 1.
! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
stopBlock
	^ self editingState pointBlock max: self editingState markBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
stopIndex
	^ self stopBlock stringIndex! !

!TextEditor methodsFor: 'mvc compatibility' stamp: 'FernandoOlivero 4/12/2011 10:41'!
storeSelectionInParagraph
	self theme currentSettings haveSelectionTextColor 
		ifTrue: [
			self text removeAttribute: TextSelectionColor primarySelection.
			self text addAttribute: TextSelectionColor primarySelection from: self startIndex to: self stopIndex - 1].
	self paragraph selectionStart: self startBlock selectionStop: self stopBlock! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
string

	^self text string! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
swapChars: aKeyboardEvent
	"Triggered byCmd-Y;.  Swap two characters, either those straddling the insertion point, or the two that comprise the selection.  Suggested by Ted Kaehler.  "

	| currentSelection aString chars |
	(chars := self selection) size = 0
		ifTrue:
			[currentSelection := self pointIndex.
			self selectMark: currentSelection - 1 point: currentSelection]
		ifFalse:
			[chars size = 2
				ifFalse:
					[morph flash. ^ true]
				ifTrue:
					[currentSelection := self pointIndex - 1]].
	aString := self selection string.
	self replaceSelectionWith: (Text string: aString reversed attributes: self emphasisHere).
	self selectAt: currentSelection + 1.
	^ true! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
text
	"Answer the text of the paragraph being edited."

	^self paragraph text! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
textStyle
	^ self paragraph textStyle! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
totalTextHeight

	^self paragraph lines last bottom! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
transformFrom: owner
	^morph transformFrom: owner! !

!TextEditor methodsFor: 'parenblinking' stamp: 'AlainPlantec 11/17/2010 18:51'!
unapplyAttribute: aTextAttribute
	"The user selected aTextAttribute to be removed.
	If there is a selection, unapply the attribute to the selection.
	In any case do not use the attribute for the user input (emphasisHere)"

	| thisSel interval |

	self editingState emphasisHere: (self emphasisHere copyWithout: aTextAttribute).
	self selection
		ifEmpty: [ self paragraph composeAll ]				
		ifNotEmpty: [:sel | self replaceSelectionWith: (sel removeAttribute: TextAttribute).
			self paragraph composeAll.
			self recomputeSelection]! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 11/16/2010 10:18'!
unapplyAttributesThat: removalBlock fromWholeParagraph: wholeParagraph
	"The user selected aTextAttribute to be removed.
	If there is a selection, unapply the attribute to the selection.
	In any case do not use the attribute for the user input (emphasisHere)
	It really looks like the 2 branches in this method should be closer!! (handling of Undo is different, and likely wrong)"

	| interval newText secondPass |

	self editingState emphasisHere: (self editingState emphasisHere reject: removalBlock).
	wholeParagraph
		ifTrue: [
			interval := self paragraph text encompassParagraph: self selectionInterval.
			newText := self paragraph text copyFrom: interval first to: interval last.
			newText
				removeAttributesThat: removalBlock
				replaceAttributesThat: [ :attr | false ]
				by: [ :attr | false ]
				signalIfWholeParagraphAttribute: false.
			self paragraph 
				replaceFrom: interval first 
				to: interval last 
				with: newText.
			self paragraph composeAll.
			self recomputeSelection ]
		
		ifFalse: [
			newText := self selection.
			newText notEmpty
				ifTrue: [
					"Do a first pass that will remove only regular attributes (i.e. isParagraphAttribute not)"
					secondPass := false.
					[
						newText
							removeAttributesThat: removalBlock
							replaceAttributesThat: [ :attr | false ]
							by: [ :attr | false ]
							signalIfWholeParagraphAttribute: true
					] on: Abort do: [ :ex |
						secondPass := true ].
					self replaceSelectionWith: newText.
					self paragraph composeAll.
					self recomputeSelection.
					"Do a second pass, calling this method again, that will remove self paragraph attributes from whole paragraphs"
					secondPass ifTrue: [
						self unapplyAttributesThat: [ :attr | attr isParagraphAttribute and: [removalBlock value: attr ]] fromWholeParagraph: true ] ] ].
	! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
undo
	"Undo  previous edit."
	self closeTypeIn.
	self editingState undo ifFalse: [morph flash]. 
! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 11/8/2010 22:09'!
undo: aKeyboardEvent 
	"Undo the last edit."

	self undo.
	^true! !

!TextEditor methodsFor: 'undoers-redoers' stamp: 'AlainPlantec 11/8/2010 22:09'!
undoRedoExchange: aninterval with: anotherInterval
	self selectInvisiblyFrom: aninterval first to: aninterval last.
	self exchangeWith: anotherInterval
! !

!TextEditor methodsFor: 'undoers-redoers' stamp: 'AlainPlantec 11/8/2010 22:09'!
undoTypeIn: aText interval: anInterval
	self selectInterval: anInterval.
	self replace: anInterval with: aText and:
		[self selectInterval: (anInterval first to: anInterval first - 1)].
! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'AlainPlantec 11/8/2010 22:09'!
unselect
	self markBlock: self pointBlock copy! !

!TextEditor methodsFor: 'private' stamp: 'CamilloBruni 2/6/2012 23:07'!
updateSelectionFromEvent: event

	| clickPoint b | 

	clickPoint := event cursorPoint.
	b := self paragraph characterBlockAtPoint: clickPoint.
	
	self editingState markBlock: b.
	self editingState pointBlock: b.
	
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'accessing' stamp: 'AlainPlantec 11/8/2010 22:09'!
visibleHeight

	^morph owner bounds height
! !

!TextEditor methodsFor: 'settings' stamp: 'AlainPlantec 11/8/2010 22:09'!
walkAlongDisplayedLine
	^ self class walkAlongDisplayedLine! !

!TextEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 11/8/2010 22:09'!
yellowButtonActivity
	"This normally opens a popup menu. Determine the selected
	item and, if one is selected, then send the corresponding message
	to either the model or the receiver."

	^ self pluggableYellowButtonActivity: false! !

!TextEditor methodsFor: 'events' stamp: 'CamilloBruni 2/6/2012 23:31'!
yellowButtonDown: anEvent
	"special case for right/yellobutton click.
	if clicking outside the current selection select the word."
	(self paragraph selectionContainsPoint: anEvent cursorPoint)
		ifFalse: [
			self mouseMove: anEvent.
			self selectWord ].! !

!TextEditor methodsFor: 'mvc compatibility' stamp: 'FernandoOlivero 3/25/2011 11:27'!
zapSelectionWith: aText
	| start stop |
	morph canChangeText ifFalse: [^ self].
	start := self startIndex.
	stop := self stopIndex.
	(aText isEmpty and: [stop > start]) ifTrue: [
		"If deleting, then set emphasisHere from 1st character of the deletion"
		self editingState emphasisHere: ((self text attributesAt: start) select: [:att | att mayBeExtended])].
	(start = stop and: [ aText size = 0 ]) ifFalse: [
		self paragraph replaceFrom: start to: stop - 1 with: aText.
		self markIndex: start; pointIndex: start + aText size.
		self editingState previousInterval: self selectionInterval selection: self selection.
		self userHasEdited  " -- note text now dirty" ].
! !
TextEditor initialize!
SmalltalkEditor initialize!
